import os 
import random

def format_script(file_path, uid=True, **kwargs):
    formatted_script = open(file_path+".format").read().format(**kwargs)
    if uid:
        exec_file_name="{}_{}{}".format(os.path.splitext(file_path)[0] ,random.random(), os.path.splitext(file_path)[1]) #
    else:
        exec_file_name = file_path
    # int "exec_file: "+exec_file_name 
    open(exec_file_name, "w+").write(formatted_script)
    return exec_file_name





import pandas as pd
import numpy as np

def remove_object_notin_network(DataSet, BioNetwork):
    index_to_remove = []
    genelist = DataSet.gene_id
    tmpb = BioNetwork.lst_g()
    # for i in genelist:
    #     if i not in tmpb:
    #         index_to_remove.append([idd for idd, ob in enumerate(genelist) if ob == i][0])

    index_to_remove = np.where(np.isin(genelist, tmpb, invert=True))[0]

    
    DataSet._remove_objects(index_to_remove)
    
    return len(index_to_remove)


def remove_nodes_notin_dataset(DataSet, BioNetwork):
    nodes_to_remove = []
    tmpg = DataSet.gene_id
    nodelist= BioNetwork.lst_g()
    # for i in nodelist:
    #     if i not in tmpg:
    #         nodes_to_remove.append(i)

    nodes_to_remove = nodelist[[np.array(np.isin(BioNetwork.lst_g(), DataSet.gene_id, invert=True))]]

    BioNetwork._removing_nodes(list(nodes_to_remove))

    return len(nodes_to_remove)

def remove_low_variance(DataSet):
    X = DataSet.ts[0]
    

    rowsums = list(map(int, np.var(X, axis=1)))
    filtered_index = []
    for x,y in enumerate(rowsums):
        if y == 0:
            filtered_index.append(int(x))
    
    DataSet._remove_objects(filtered_index)

    return(len(filtered_index))

def filter_with_cutoff(DataSet, cutoff):
    X = DataSet.ts[0]

    rowsums = np.mean(X, axis=1)
    if cutoff==0:
        filtered_index = np.where(rowsums==cutoff)[0]
    else:
        filtered_index = np.where(rowsums<cutoff)[0]
    
    filtered_genes = [DataSet.gene_id[x] for x in filtered_index]
    DataSet._remove_objects(filtered_index)
    return(len(filtered_index), filtered_genes)


def preprocess(DataSet, BioNetwork=None, remove_low_var=False, cutoff=0):

    """Preprocess data, remove objects without expression along all timepoints.

    Parameters
    ----------

    DataSet : Dataset object.

    BioNetwork : BioNetwork object. If provided, objects that is not in the network will be removed.

    remove_low_var : (boolean) default=False. If true, objects with variance 0 will be removed.

    cutoff : default=0. If given, objects will mean expression across all timepoints lower than the cutoff will be removed.

    Returns 
    ------
    None, changes made directly in the DataSet object


    """
    print(f"Input data dimension: {DataSet.timeserieslist.shape}")
    
    x, filtered_genes = filter_with_cutoff(DataSet, cutoff)
    if cutoff >0:
        print(f"Removed {x} objects lower than {cutoff}")
    else:
        print(f"Removed {x} with 0 values.")
    
    if remove_low_var:
        y= remove_low_variance(DataSet)
        print("Removed {} objects with 0 variance.".format(y))
    
    if BioNetwork is not None:
        i = remove_nodes_notin_dataset(DataSet, BioNetwork)
        print("Removed {} objects from dataset that are not in the network".format(i))
        j = remove_object_notin_network(DataSet, BioNetwork)
        print("Removed {} nodes from network that are not in the dataset (included the genes with lower than cutoff expression).".format(j))

    print("Filtered data: {}".format(DataSet.timeserieslist.shape))





import numpy as np
from sklearn.cluster import *
from sklearn.metrics.cluster import adjusted_rand_score
from sklearn.metrics import pairwise_distances, silhouette_score
from tslearn.metrics import cdist_soft_dtw
from tslearn.clustering import TimeSeriesKMeans
from tslearn.clustering import silhouette_score as tssilhouette_score
from sklearn.metrics import davies_bouldin_score
from sklearn.metrics import pairwise_distances, silhouette_score

PROTOTYPE = {'median':np.median, 'mean':np.mean}

def intra_clusters_similarity(timeserieslist, clusters, prototype, metric):
    ##get array of gene in order of clustering
    a = [y for x in clusters.index_clusters.values() for y in x]
    ##get array of cluster in order of clustering
    b = [u for u,v in clusters.index_clusters.items() for _ in v]
    ## arg sort the gene array
    ## map the cluster array with arg sorted gene array
    cluster_array = [b[i] for i in np.argsort(a)]
    ## calculate similarity the time series object with corresponding prototype
    prototype_arr = [clusters._prototype[x] for x in cluster_array]

    assert len(prototype_arr) == timeserieslist.shape[1]
    if metric == "soft_dtw":
        proto_cluster_sim_arr = [cdist_soft_dtw([prototype_arr[x], np.squeeze(timeserieslist.tiu_timeserieslist[:,x,:],axis=0)]) for x in range(len(prototype_arr))]
    else:
        proto_cluster_sim_arr = [pairwise_distances([prototype_arr[x], np.squeeze(timeserieslist.tiu_timeserieslist[:,x,:],axis=0)], metric="euclidean")[0,1] for x in range(len(prototype_arr))]

    ## calculate average similarities of the clusters 
    all_clu_sim = []
    for clu in clusters.index_clusters.keys():
        this_clu = [x==clu for x in cluster_array]
        avg_sim = 0
        for sim_clu in range(len(this_clu)):
            if this_clu[sim_clu]:
                avg_sim+=proto_cluster_sim_arr[sim_clu]
        avg_sim = avg_sim/np.sum(this_clu)
        all_clu_sim.append(avg_sim)
    
    return all_clu_sim


import networkx as nx
from collections import defaultdict
from scipy.stats import chi2
import numpy as np
import pandas as pd
from functools import reduce

from .DOMINO.src.core import domino
from .DOMINO.src.core import preprocess_slices as sl
from .DataSet import DataSet
from .clustering import clustering

def _combine_pvals(pvals):
        chisqprob = lambda chisq, df: chi2.sf(chisq, df)
        s = -2 * np.sum(np.log(pvals))
        
        return chisqprob(s, 2 * len(pvals))

def _check_nodes(target, network_file):
    if type(network_file)==str:
        G = nx.read_edgelist(network_file)
    else:
        G = network_file.networkz
    
    checkedtarget = []
    for i in target:
        if i in set(G.nodes):
            checkedtarget.append(i)

    return checkedtarget

def _affected_domains(targets, ascov):
    # #get genes switching events
    DDInodes=[]

    for g in ascov.iterrows():
        for domain in g[1]['exclusive_domains']:
            tmp = str(g[1]['gene']+"/"+str(domain))
    
            DDInodes.append(tmp)
    
    return DDInodes
  
            
            
def run_domain_domino(target, is_results, name=None, scores = None, network_file = "data/network/network_human_PPIDDI.tab", output_file_path = "./slices/slices.txt", 
                run_cluster=None, slice_threshold=0.3, module_threshold=0.05, prize_factor = 0, n_steps=20):
    '''
    Parameters
    -----------
    target :
        clustering object from spycone or gene list in entrez ID
    is_results : DataFrame
        Data Frame of isoform switch detection result
    scores : None
        activity scores of the genes (e.g. p-values from differential expression analysis)

    run_cluster :
        Specify the cluster name if you only want to run a specific cluster

    Network file : str
        default: "data/network/network_human_PPIDDI.tab"

    output file path
        default: output slices file for DOMINO.

    slice_threshold : float
    
    module_threshold : float

    prize_factor : float 

    n_steps : int
    '''
    if name is None:
        name="a"

    if is_results is not None:
        genescores = defaultdict(list)
        for i in range(is_results.shape[0]):
            genescores[is_results.gene.tolist()[i]].append(is_results.adj_pval.tolist()[i])

        for u,v in genescores.items():
            if len(v)>1:
                genescores[u] = _combine_pvals(v)
            else:
                genescores[u] = v[0]
    else:
        genescores = None
    

    sl.create_slices(network_file, output_file_path)
    if isinstance(target, list):
        ##check list    

        DDInodes = _affected_domains(list(map(str,target)), is_results)
        checkedtarget = _check_nodes(DDInodes, network_file)

        ##TODO scores

        ## clusterobj can also be list
        a = defaultdict(list)
        
        tmp, scores = domino.main(DDInodes, network_file, slices_file=output_file_path, slice_threshold=slice_threshold,  module_threshold=module_threshold, prize_factor=prize_factor, n_steps=n_steps)
        a[name].append(tmp)
        a[name].append(scores)

        return a

    elif isinstance(target, clustering) and run_cluster is not None:
        a = defaultdict(list)
        gene_list = []
        for cc in run_cluster:
            gene_list.append(target.genelist_clusters[cc])
        
        gene_list = reduce(lambda x,y: x+y, gene_list)
        checkedtarget = _check_nodes(list(map(str,gene_list)), network_file)

        DDInodes = _affected_domains(checkedtarget, is_results) ##get affected domains
        tmp, scores = domino.main(DDInodes, network_file, slices_file=output_file_path, slice_threshold=slice_threshold,  module_threshold=module_threshold, prize_factor=prize_factor, n_steps=n_steps)

        a[','.join(map(str,run_cluster))].append(tmp)
        a[','.join(map(str,run_cluster))].append(scores)

        return a

    else:

        a=defaultdict(list)
        for u,v in target.genelist_clusters.items():
            

            
            #scoresdf.to_csv("/nfs/home/students/chit/lrz_ticone/domino_emp/{}_cluster{}_mod.csv".format(name, u), index=False)
            ### 
            checkedtarget = _check_nodes(list(map(str,v)), network_file)
            DDInodes = _affected_domains(checkedtarget, is_results)

            scores = []
            for gene in DDInodes:
                if genescores is not None:
                    if gene in genescores.keys():
                        scores.append(genescores[gene])
                    else:
                        scores.append(1)

                else:
                    scores.append(1)
            ##fornow emp
            scoresdf = pd.DataFrame(index=list(map(str,v)), dtype=str)
            ## 
            ##TODO scores

            ddimod, scores=domino.main(DDInodes, network_file=network_file, slices_file=output_file_path, slice_threshold=slice_threshold,  module_threshold=module_threshold, prize_factor=prize_factor, n_steps=n_steps)
            
            a[u].append(ddimod)
            a[u].append(scores)

        print("---------Network enrichment Result---------\n")
        for u,v in a.items():
            for e, vv in enumerate(v[0]):
                print(f"Cluster {u} Module {e} has {len(vv)} nodes.")
        print("-----END-----")        
        return a 


    


def run_domino(target, name=None, is_results=None, scores = None, network_file = "data/network/mouse_biogrid_entrez.tab", output_file_path = "./slices/slices.txt", 
                run_cluster=None, slice_threshold=0.3, module_threshold=0.05, prize_factor = 0, n_steps=20):
    '''
    Parameters
    -----------
    target :
        clustering object from spycone or gene list in entrez ID
    is_results : DataFrame
        Data Frame of isoform switch detection result
    scores : None
        activity scores of the genes (e.g. p-values from differential expression analysis)

    run_cluster :
        Specify the cluster name if you only want to run a specific cluster

    Network file : str
        default: "data/network/network_human_PPIDDI.tab"

    output file path
        default: output slices file for DOMINO.
        
    slice_threshold : float
    
    module_threshold : float

    prize_factor : float 

    n_steps : int
    '''
    print("start running DOMINO...")
    if name is None:
        name="a"

    if is_results is not None:
        genescores = defaultdict(list)
        for i in range(is_results.shape[0]):
            genescores[is_results['gene_symb'].tolist()[i]].append(is_results.adj_pval.tolist()[i])

        for u,v in genescores.items():
            if len(v)>1:
                genescores[u] = _combine_pvals(v)
            else:
                genescores[u] = v[0]
    else:
        genescores = None
    

    sl.create_slices(network_file, output_file_path)
    # if isinstance(target, list):
    #     ##check list    
    #     checkedtarget = _check_nodes(list(map(str,target)), network_file)

    #     ## clusterobj can also be list
    #     a = defaultdict(list)

    #     tmp, scores = domino.main(list(map(str,checkedtarget)), network_file, slices_file=output_file_path, slice_threshold=slice_threshold,  module_threshold=module_threshold, prize_factor=prize_factor, n_steps=n_steps)
    #     a[name].append(tmp)
    #     a[name].append(scores)

    #     return a

    # elif isinstance(target, clustering) and run_cluster is not None:
    #     a = defaultdict(list)
    #     gene_list = []
    #     for cc in run_cluster:
    #         gene_list.append(target.genelist_clusters[cc])
        
    #     gene_list = reduce(lambda x,y: x+y, gene_list)
    #     checkedtarget = _check_nodes(list(map(str,gene_list)), network_file)

    #     tmp, scores = domino.main(list(map(str, checkedtarget)), network_file, scores=scores, slices_file=output_file_path, slice_threshold=slice_threshold,  module_threshold=module_threshold, prize_factor=prize_factor, n_steps=n_steps)

    #     a[','.join(map(str,run_cluster))].append(tmp)
    #     a[','.join(map(str,run_cluster))].append(scores)

    #     return a

    # else:

    a=defaultdict(list)
    for u,v in target.genelist_clusters.items():
        scores = []
        for gene in target.symbs_clusters[u]:
            if genescores is not None:
                if gene in genescores.keys():
                    scores.append(genescores[gene])
                else:
                    scores.append(1)

            else:
                scores.append(1)
        ##fornow emp
        ##TODO scores
        
        #scoresdf.to_csv("/nfs/home/students/chit/lrz_ticone/domino_emp/{}_cluster{}_mod.csv".format(name, u), index=False)
        ### 
        checkedtarget = _check_nodes(list(map(str,v)), network_file)
        
        tmp, scores = domino.main(list(map(str,checkedtarget)), network_file,  slices_file=output_file_path, slice_threshold=slice_threshold,  module_threshold=module_threshold, prize_factor=prize_factor, n_steps=n_steps)
        a[u].append(tmp)
        a[u].append(scores)

    print("---------Network enrichment Result---------\n")
    for u,v in a.items():
        #for e, vv in enumerate(v[0]):
        print(f"Cluster {u} found {len(v[0])} module(s).")
    print("-----END-----")        
    return a 



import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pandas as pd
import scipy.stats

from ..inputdata.DataSet import DataSet
from ..inputdata.BioNetwork import BioNetwork
#from stat.pq_val import _cal_pqvals

def _plot_mean_and_CI(mean, lb, ub, tps, clus, color_mean=None, color_shading=None,  xlab="Time points", ylab="Gene expression"):
    # plot the shaded range of the confidence intervals
    plt.fill_between(range(mean.shape[0]), ub, lb,
                    color=color_shading, alpha=.5)
    # plot the mean on top
    plt.plot(mean, color_mean)
    plt.title('Mean 95% confidence interval of Cluster {}'.format(clus))
    #plt.ylabel("Hyper-similarity score")
    plt.ylabel(ylab)
    plt.xlabel(xlab)
    plt.xticks(np.arange(tps),np.arange(start=1,stop=tps+1,step=1))

def _mean_confidence_interval(data, confidence):
    a = np.array(data)
    n = len(a)
    m, se = np.mean(a), scipy.stats.sem(a)
    h = se * scipy.stats.t.ppf((1 + confidence) / 2., n-1)
    return m, m-h, m+h

def plotting_clusters(DataSet, confidence=0.95, xlab="Time points", ylab="Gene expression", plot="none"):
    all_g_names = {}
    wilc = []
    t1 = DataSet.ts[1]
    l = DataSet.l[0]
    try:
        t2 = DataSet.group2[1]
    except:
        t2 = None
    object_index = DataSet.object_index[0]
    use_clu = DataSet.use_clu[0]
    symbs1 = DataSet.symbs[0]
    gene_list = DataSet.gene_list[0]
    tps = DataSet.timepts

    if t2 ==None:
        col = ["red", "blue", "green", "yellow", "purple", "white", "brown", "orange", "teal", "silver", "pink"]

        for ii, i in use_clu.items():
            if len(i) != 0:#len(nms):
                if len(i) >= l:
                    tmm = {}
                    for rind in i:
                        tmm[str(gene_list[rind])] = symbs1[rind]
                    all_g_names[str(ii)] = tmm

                    ub1 = []
                    lb1 = []
                    m1 = []
                    for ar in t1[list(i)].T:
                        x,y,z = _mean_confidence_interval(ar, confidence=0.95)
                        m1.append(x)
                        lb1.append(y)
                        ub1.append(z)
                    m1 = np.array(m1)
                    ub1 = np.array(ub1)
                    lb1 = np.array(lb1)

                    wilc.append([t1[list(i)]])
                    # fig = plt.figure(1, figsize=(12, 10))
                    # plot_mean_and_CI_ylim(m1, ub1, lb1, color_mean='red', color_shading='lightcoral')
                    # plt.savefig(other_path + "/" + str(j) +"_"+ args.name + "_red_{}_ylim_conf.pdf".format(j)) 
                    # plt.close()
                    
                    if plot != "all":
                        if type(plot) !="list":
                            plot = [plot]
                        if ii in plot:
                            fig = plt.figure(1, figsize=(12, 10))
                            _plot_mean_and_CI(m1, ub1, lb1, tps, clus=ii, color_mean='red', color_shading='lightcoral', xlab=xlab, ylab=ylab)
                            #plt.savefig(other_path + "/" + str(j) +"_"+ args.name + "_red_{}_conf.png".format(j)) 
                            #plt.savefig(other_path + "/" + str(j) +"_"+ args.name + "_red_{}_conf.pdf".format(j)) 
                            plt.show()
                        else:
                            continue

    
    if t2 != None:
        if len(t2)>0:
            #plt.scatter(d1[:,0], d1[:,1], c="gray")
            col = ["red", "blue", "green", "yellow", "purple", "white", "brown", "orange", "teal", "silver", "pink"]
            j = -1
            for ii, i in use_clu.items():
                if len(i) != 0:#len(nms):
                    #if list(labls).index("3040") in i:
                    #    print("XXXXXXXXXXXXXXXXXXX", i)
                    if len(i) >= l: 
                        j += 1

                        tmm = {}
                        for rind in i:
                            tmm[str(gene_list[rind])] = symbs1[rind]
                        all_g_names[str(ii)] = tmm

                        ub1 = []
                        lb1 = []
                        m1 = []
                        for ar in t1[list(i)].T:
                            x,y,z = _mean_confidence_interval(ar, confidence=0.95)
                            m1.append(x)
                            lb1.append(y)
                            ub1.append(z)
                        m1 = np.array(m1)
                        ub1 = np.array(ub1)
                        lb1 = np.array(lb1)

                        ub2 = []
                        lb2 = []
                        m2 = []

                        for ar in t2[list(i)].T:
                            x,y,z = _mean_confidence_interval(ar, confidence=0.95)
                            m2.append(x)
                            lb2.append(y)
                            ub2.append(z)
                        m2 = np.array(m2)
                        ub2 = np.array(ub2)
                        lb2 = np.array(lb2)

                        wilc.append([t1[list(i)], t2[list(i)]]) #for calculating wilcoxonstats


                        #print([M1[i].flatten(), M2[i].flatten()])
                        
                        if plot != "none":
                            if type(plot)!="list":
                                plot = [plot]
                            if ii in plot:
                                fig = plt.figure(1, figsize=(12, 10))
                                _plot_mean_and_CI(m1, ub1, lb1, tps, clus=ii, color_mean='red', color_shading='lightcoral', xlab=xlab, ylab=ylab)
                                _plot_mean_and_CI(m2, ub2, lb2, tps, clus = ii, color_mean='green', color_shading='lightgreen', xlab=xlab, ylab=ylab)
                                #plt.savefig(other_path + "/" + str(j) +"_"+ args.name + "_red_green_{}_conf.pdf".format(j)) ## png
                                #plt.savefig(other_path + "/" + str(j) +"_"+ args.name + "_red_green_{}_conf.png".format(j)) ## png
                                plt.show()
                            else:
                                continue


    
    DataSet.add_for_graph(all_g_names, wilc)

def _make_simple_graph_plot(gtt1, edg, col, fsize=(40, 40), k=0.2, iterations=90):

        plt.figure(1, figsize=fsize)
        pos = nx.spring_layout(gtt1, seed=42, k=k, iterations=iterations)
        #nx.draw_networkx_nodes(gtt1, pos, nodelist=gtt1.nodes(), node_color="grey",node_size=1000)
        nx.draw_networkx_nodes(gtt1, pos, nodelist=gtt1.nodes(), node_color=col,node_size=1500)


        nx.draw_networkx_edges(gtt1, pos, alpha=0.5, width=3, edge_color="grey")
        nx.draw_networkx_labels(gtt1, pos, font_size=50, font_family='sans-serif', font_color="k")
        plt.axis('off')
        #plt.savefig(name)
        plt.show()

def extract_subnetworks(DataSet, BioNetwork, plot="all"):
    _cal_pqvals(DataSet)

    object_entrez = DataSet.all_g_names[0]
    mtrx = DataSet.mtrx[0]
    object_simval = DataSet.object_simval[0]
    l = DataSet.l[0]
    clu_obj = DataSet.use_clu[0]
    pvals = DataSet.pval[0]

    g = BioNetwork.g()
    mtrx_av = np.average(mtrx)
    print("making graphs")
    #### for bias graph
    
    if plot=="all":
        x=0
        for i,j in object_entrez.items():
            if len(j) >= l:
                print("p-values for cluster {} : {}".format(i, pvals.values()[x]))
                gtmp = g.subgraph(list(j.keys()))
                gtmp = nx.relabel_nodes(gtmp, j)

                _make_simple_graph_plot(gtmp, gtmp.edges(), col="red")
                x+=1

        print("Done making graphs.")
    else:
        if type(plot)!="list":
            plot = [plot]
        for x in plot:
            j = object_entrez[str(x)]
            print("p-values for cluster {} : {}".format(int(x), pvals[x]))
            gtmp = g.subgraph(list(j.keys()))
            gtmp = nx.relabel_nodes(gtmp, j)

            _make_simple_graph_plot(gtmp, gtmp.edges(), col="red")
    


"""
Implement linear threshold models
"""
#!/usr/bin/env python
#    Copyright (C) 2004-2010 by
#    Hung-Hsuan Chen <hhchen@psu.edu>
#    All rights reserved.
#    BSD license.
#    NetworkX:http://networkx.lanl.gov/.
__author__ = """Hung-Hsuan Chen (hhchen@psu.edu)"""

import sys
sys.setrecursionlimit(100000000)
import copy
import networkx as nx

__all__ = ['linear_threshold']

def linear_threshold(G, seeds, steps=0):
  """Return the active nodes of each diffusion step by linear threshold model
  Parameters
  ----------
  G : networkx graph
      The number of nodes.
  seeds: list of nodes
      The seed nodes of the graph
  steps: int
      The number of steps to diffuse
      When steps <= 0, the model diffuses until no more nodes
      can be activated
  Return
  ------
  layer_i_nodes : list of list of activated nodes
    layer_i_nodes[0]: the seeds
    layer_i_nodes[k]: the nodes activated at the kth diffusion step
  Notes
  -----
  1. Each node is supposed to have an attribute "threshold".  If not, the
     default value is given (0.5).
  2. Each edge is supposed to have an attribute "influence".  If not, the
     default value is given (1/in_degree)
  References
  ----------
  [1] GranovetterMark. Threshold models of collective behavior.
      The American journal of sociology, 1978.
  Examples
  --------
  >>> DG = nx.DiGraph()
  >>> DG.add_edges_from([(1,2), (1,3), (1,5), (2,1), (3,2), (4,2), (4,3), \
  >>>   (4,6), (5,3), (5,4), (5,6), (6,4), (6,5)])
  >>> layers = networkx_addon.information_propagation.linear_threshold(DG, [1])
  """
  if type(G) == nx.MultiGraph or type(G) == nx.MultiDiGraph:
      raise Exception( \
          "linear_threshold() is not defined for graphs with multiedges.")

  # make sure the seeds are in the graph
  for s in seeds:
    if s not in G.nodes():
      raise Exception("seed", s, "is not in graph")

  # change to directed graph
  if not G.is_directed():
    DG = G.to_directed()
  else:
    DG = copy.deepcopy(G)

  # init thresholds
  for n in DG.nodes():
    if 'threshold' not in DG.nodes[n]:
      DG.nodes[n]['threshold'] = 0.5
    elif DG.nodes[n]['threshold'] > 1:
      raise Exception("node threshold:", DG.nodes[n]['threshold'], \
          "cannot be larger than 1")

  # init influences
  in_deg = DG.in_degree()
  for e in DG.edges():
    if 'influence' not in DG[e[0]][e[1]]:
      DG[e[0]][e[1]]['influence'] = 1.0 / in_deg[e[1]]
    elif DG[e[0]][e[1]]['influence'] > 1:
      raise Exception("edge influence:", DG[e[0]][e[1]]['influence'], \
          "cannot be larger than 1")

  # perform diffusion
  A = copy.deepcopy(seeds)
  if steps <= 0:
    # perform diffusion until no more nodes can be activated
    return _diffuse_all(DG, A)
  # perform diffusion for at most "steps" rounds only
  return _diffuse_k_rounds(DG, A, steps)

def _diffuse_all(G, A):
  layer_i_nodes = [ ]
  layer_i_nodes.append([i for i in A])
  while True:
    len_old = len(A)
    A, activated_nodes_of_this_round = _diffuse_one_round(G, A)
    layer_i_nodes.append(activated_nodes_of_this_round)
    if len(A) == len_old:
      break
  return layer_i_nodes

def _diffuse_k_rounds(G, A, steps):
  layer_i_nodes = [ ]
  layer_i_nodes.append([i for i in A])
  while steps > 0 and len(A) < len(G):
    len_old = len(A)
    A, activated_nodes_of_this_round = _diffuse_one_round(G, A)
    layer_i_nodes.append(activated_nodes_of_this_round)
    if len(A) == len_old:
      break
    steps -= 1
  return layer_i_nodes

def _diffuse_one_round(G, A):
  activated_nodes_of_this_round = set()
  for s in A:
    nbs = G.successors(s)
    for nb in nbs:
      if nb in A:
        continue
      active_nb = list(set(G.predecessors(nb)).intersection(set(A)))
      if _influence_sum(G, active_nb, nb) >= G.nodes[nb]['threshold']:
        activated_nodes_of_this_round.add(nb)
  A.extend(list(activated_nodes_of_this_round))
  return A, list(activated_nodes_of_this_round)

def _influence_sum(G, froms, to):
  influence_sum = 0.0
  for f in froms:
    influence_sum += G[f][to]['influence']
  return influence_sum


import pandas as pd
import numpy as np
import time
from scipy import stats
from statsmodels.stats.multitest import multipletests

from ..DataSet import DataSet
from ..BioNetwork import BioNetwork

def _cal_pqvals(DataSet):
    clu_val = DataSet.use_clu[0]
    graph_nms = DataSet.object_index[0]
    clst_values = DataSet.object_simval[0]
    mtrx = DataSet.mtrx[0]
    l=DataSet.l[0]
    wilc = DataSet.wilc[0]

    pvals = {}
    mtrx_av = np.average(mtrx)
    assert len(graph_nms) == len(clst_values)
    for i in range(len(graph_nms)):
        if len(graph_nms[i]) >= l:
            pvals[clu_val.keys()[i]] = (stats.mannwhitneyu(clst_values[i], mtrx_av.flatten(), alternative="two-sided")[1] )
        


    nr_of_pvs = len(pvals)
#     for i in wilc:
#             pvals.append(stats.wilcoxon(np.average(i[0],axis=0), np.average(i[1],axis=0) )[1])

    qvals_true = multipletests(pvals.values(), method="fdr_bh")[1]

    qvals = qvals_true[:nr_of_pvs]
    p_gr = pvals.items()[:nr_of_pvs] 
    p_gr = dict(p_gr)
    DataSet.add_pval(p_gr)
    DataSet.add_qval(qvals)


from setuptools import setup, Extension
from Cython.Build import cythonize
from Cython.Distutils import build_ext
from numpy.distutils.misc_util import Configuration, get_numpy_include_dirs
import numpy

extensions = [
    Extension("connectivity_task", 
    ["connectivity_task.pyx"],
    include_dirs = [numpy.get_include()]
    )
]

setup(
    ext_modules = cythonize(extensions, gdb_debug=True),
    cmdclass = {'build_ext':build_ext}
    
)



import pandas as pd
import numpy as np
import time
import random
import warnings
import math
import itertools as it
from sklearn.metrics import davies_bouldin_score

from ..DataSet import DataSet
from ..BioNetwork import BioNetwork
from .._clustering.clusterobj import clusterObj
from .._clustering import similarities as sim


def dbindex(clusterObj, simfunc="pearson"):
    alltimeseries = np.array(clusterObj.DataSet.timeserieslist, dtype="double")
    clusters = clusterObj.index_clusters
    prototype= clusterObj._prototype
    simfunc = clusterObj.similarityfunction
    #average dist/corr between each cluster and its medoid
    avg_per_clusters = sim.intra_clusters_similarity(alltimeseries, clusters, prototype, simfunc)

    #compute distance of each clusters pairs
    clusters_pair_sim = sim.clusters_pair_similarity(clusterObj._prototype, simfunc)

    #sum of average dist of each clusters pairs / their distances --> get the max
    values=np.zeros(len(clusters))
    clist = list(clusters.keys())

    dbi=0
    n=0
    for o in range(len(clist)):
        k=clist[o]
        m=0
        for l in clist[0:o]+clist[o+1:]:
            values[o] = (avg_per_clusters[o]+avg_per_clusters[m])/clusters_pair_sim[n][m]

            
            m+=1
        dbi += np.max(values)
        n+=1

    result = dbi/len(clusters)
    
    
    return result


import os
import numpy as np
import multiprocessing

USE_CACHE=True
N_OF_THREADS=5 # int(np.ceil(multiprocessing.cpu_count()*0.9))
dir_path = os.path.dirname(os.path.realpath(__file__))
PATH_TO_CONF = "env/config/conf.json"

REPO_DIR = os.path.dirname(os.path.realpath(__file__))
SH_DIR = os.path.join(REPO_DIR, "sh","scripts")


LABEL_ID = "sample_type.samples"
PRIMARY_TUMOR = "Primary Tumor"
METASTATIC = "Metastatic"

LABELS_NORMAL = "labels_normal"
LABELS_SHUFFLE = "labels_shuffle"
LABELS_RANDOM = "labels_random"
LABELS_ALTERNATED = "labels_alternated"
LABELS_INVERTED = "labels_inverted"

ENSG_TO_GENE_SYMBOLS = "ensg2gene_symbol.txt"
ENSMUSG_TO_GENE_SYMBOLS = "ensmusg2gene_symbol.txt"
ENSEMBL_TO_ENTREZ = "ensembl2entrez.txt"

GO_OBO_URL = 'http://purl.obolibrary.org/obo/go/go-basic.obo'
GO_ASSOCIATION_GENE2GEO_URL = 'https://ftp.ncbi.nlm.nih.gov/gene/DATA/gene2go.gz'
GO_FILE_NAME = 'go_bp.obo' #'go-basic.obo'
GO_ASSOCIATION_FILE_NAME = "gene2go"




import pandas as pd
import numpy as np
import time
import random
import warnings
from joblib import Parallel, delayed

from ..DataSet import DataSet
from ..BioNetwork import BioNetwork
from .._clustering.similarities import intra_clusters_similarity
from .._prototype.prototype_builder import _prototype_builder


##features "number of objects", "similarity value"
class featuresObj():
    def __init__(self, testobject, timeserieslist, feature_type=None, feature_store = None, features=["prototype_standard_variance","average_similarity","number_of_object"], seed=1234):
        self.features = features #prototypesv, averagesim, noofobj
        self.testobject = testobject
        self.timeserieslist = timeserieslist #np.array
        self.feature_store = feature_store ##np.array
        self.feature_type = feature_type #connectivity /clusters / shuffle / iso_pairs / shuffle_iso_pairs
        self.seed = seed
        
        if feature_store is None:
            if feature_type == "clusters" or feature_type == "shuffle":
                try:
                    self.get_clusters_features()
                except:
                    raise ValueError("Clustering object is needed for clusters feature.")

            if feature_type == "iso_pairs":
                #test object data frame of isopairs id
                self.get_switching_features()
            
            elif feature_type == "shuffle_iso_pairs":
                self.get_shuffle_switching_features()

    
    def _cluster_prototype_standard_variance(self):
        if self.feature_type == "clusters":
            prototypes = self.testobject.clusterobj._prototype
            values = []
            for index,i in prototypes.items():
                values.append(np.var(i))
            return values

        if self.feature_type == "shuffle":
            
            clusters = self.testobject.clusterobj.index_clusters
            timeseries_object = self.timeserieslist
            prototype_func = self.testobject.clusterobj.prototypefunction

            prototypes = _prototype_builder(timeseries_object, clusters, prototype_func=prototype_func)
            values = []
            for index,i in prototypes.items():
                values.append(np.var(i))
            return 1/values

    def _cluster_average_similarity(self) -> np.ndarray:
        if self.feature_type == "clusters":
            values = intra_clusters_similarity(self.testobject.timeserieslist, self.testobject.clusterobj.index_clusters, self.testobject.clusterobj._prototype, self.testobject.clusterobj.metric)

            values = np.array(values)
            return values

        if self.feature_type == "shuffle":
            timeseries_object = self.timeserieslist
            clusters = self.testobject.clusterobj.index_clusters

            all_score = intra_clusters_similarity(timeseries_object, clusters, self.testobject.clusterobj._prototype, self.testobject.clusterobj.metric)

            values = all_score
            return values

    def _cluster_information_content(self):
        pass

    def _cluster_number_of_objects(self) -> np.ndarray:
        if self.feature_type == "clusters":
            clusterobj = self.testobject.clusterobj.genelist_clusters
            values = []
            
            for u, v in clusterobj.items():
                values.append(len(v))

            values = np.array(values)
            return values

        if self.feature_type == "shuffle":
            clusterobj = self.clusterobj.genelist_clusters
            values = []
            
            for u, v in clusterobj.items():
                values.append(len(v))

            values = np.array(values)
            return values

    def _clustering_total_obj_cluster_similarity(self):
        pass

    def _clustering_number_of_clusters(self):
        pass

    def _take_interval(self, sp):
            #check interval between switch points
        newsp = np.insert(sp,0,0)
        newsp = np.append(newsp, self.testobject.timepts)
        intervals = []
        for i in range(len(sp)):
            intervals.append(newsp[i+2]-newsp[i+1])
        return intervals

    def _iso_pairs_switch_ratio(self):
        #return iso switching ratio for each pairs
        isopairs = self.testobject.isoobj.isopairs
        nisopairs = len(self.testobject.isoobj.isopairs['major_transcript'])
        iso_ratio = np.zeros(nisopairs)
        iso_dict_info = self.testobject.isoobj.iso_dict_info

        def _cal_iso_ratio(n):
        # for n in range(nisopairs):
            iso1 = iso_dict_info[isopairs['major_transcript'][n]]
            iso2 = iso_dict_info[isopairs['minor_transcript'][n]]
            

            try:
                arr1 = self.testobject.isoobj.normdf[iso1[0]]['normarr'][:,iso1[1],:]
                arr2 = self.testobject.isoobj.normdf[iso2[0]]['normarr'][:,iso2[1],:]

                orgarr1 = self.testobject.isoobj.normdf[iso1[0]]['array'][:,iso1[1],:]
                orgarr2 = self.testobject.isoobj.normdf[iso2[0]]['array'][:,iso2[1],:]
            except:
                raise ValueError("{},{} doesn't seem valid.".format(iso1,iso2))
           
            ir, alldiff, final_sp, cor = self.testobject.isoobj._iso_switch_between_arrays(arr1,arr2,orgarr1,orgarr2)

            if any(alldiff):
                bp = np.argmax(alldiff)
                tmp_sw=np.insert(final_sp, 0,0)
                tmp_sw=np.append(tmp_sw, self.testobject.timepts)
                iso_inter = np.max(self._take_interval(alldiff))
                #iso_prob = ((np.sum(arr1[:,tmp_sw[bp-1]:tmp_sw[bp]].reshape(1,-1)[0]>arr2[:,tmp_sw[bp-1]:tmp_sw[bp]].reshape(1,-1)[0]))/arr1[:,tmp_sw[bp-1]:tmp_sw[bp]].reshape(1,-1)[0].shape[0]) + ((np.sum(arr1[:,tmp_sw[bp]:tmp_sw[bp+1]].reshape(1,-1)[0]<arr2[:,tmp_sw[bp]:tmp_sw[bp+1]].reshape(1,-1)[0]))/(arr1[:,tmp_sw[bp]:tmp_sw[bp+1]].reshape(1,-1)[0].shape[0]))
                dv = np.max(alldiff)
            else:
                dv = 0
                iso_prob=0
                iso_inter=0
                
            return dv, cor, iso_inter
        #
        #parallel run 
        res = Parallel(backend="loky") (delayed(_cal_iso_ratio)(n) for n in range(nisopairs))
        diff, corr, iso_inter = map(list,zip(*res))
        return corr, iso_inter

    def _shuffle_iso_pairs_switch_ratio(self):
        #return iso switching ratio for each pairs
        isopairs = self.testobject.isoobj.isopairs #id of iso pairs
        nisopairs = len(isopairs['major_transcript'])
        iso_dict_info = self.testobject.isoobj.iso_dict_info

        #make normdf for shuffle
        warnings.simplefilter("ignore")
        from collections import defaultdict
        from functools import partial
        
        groupdict = defaultdict(lambda: defaultdict(list))
        for idx, sym in enumerate(self.testobject.gene_id):
            groupdict[sym]['indices'].append(idx)
            groupdict[sym]['array'].append(self.timeserieslist[:,idx,:])
    
        
        #normlize work
        for sym, df in groupdict.items():
            df['array'] = np.concatenate(df['array']).reshape(self.testobject.reps1,len(df['array']),self.testobject.timepts)
            tmp = df['array']/np.sum(df['array'], axis=1)
            tmp[np.isnan(tmp)] = 0
            for i in range(tmp.shape[0]):
                #np.random.shuffle(tmp[i,:,:])
                for j in range(tmp.shape[1]):
                    np.random.seed(self.seed)
                    np.random.shuffle(tmp[i,j,:])

            groupdict[sym]['normarr'] = tmp

        def _cal_iso_ratio(n):
        # for n in range(nisopairs):
            iso1 = iso_dict_info[isopairs['major_transcript'][n]]
            iso2 = iso_dict_info[isopairs['minor_transcript'][n]]

            arr1 = groupdict[iso1[0]]['normarr'][:,iso1[1],:]
            arr2 = groupdict[iso2[0]]['normarr'][:,iso2[1],:]

            # orgarr1 = groupdict[iso1[0]]['array'][:,iso1[1],:]
            # orgarr2 = groupdict[iso2[0]]['array'][:,iso2[1],:]

            ir, alldiff, final_sp, cor = self.testobject.isoobj._iso_switch_between_arrays(arr1, arr2, arr1, arr2)
            
            if any(alldiff):
                bp = np.argmax(alldiff)
                tmp_sw=np.insert(final_sp, 0,0)
                tmp_sw=np.append(tmp_sw, self.testobject.timepts)
                iso_inter = self._take_interval(alldiff)
                #iso_prob = (((np.sum(arr1[:,tmp_sw[bp-1]:tmp_sw[bp]].reshape(1,-1)[0]>arr2[:,tmp_sw[bp-1]:tmp_sw[bp]].reshape(1,-1)[0]))/arr1[:,tmp_sw[bp-1]:tmp_sw[bp]].reshape(1,-1)[0].shape[0]) + ((np.sum(arr1[:,tmp_sw[bp]:tmp_sw[bp+1]].reshape(1,-1)[0]<arr2[:,tmp_sw[bp]:tmp_sw[bp+1]].reshape(1,-1)[0]))/(arr1[:,tmp_sw[bp]:tmp_sw[bp+1]].reshape(1,-1)[0].shape[0])))/2
                return np.max(alldiff), cor, np.max(iso_inter)
            else:
                return 0, cor, 0

        res = Parallel(backend="loky") (delayed(_cal_iso_ratio)(n) for n in range(nisopairs))

        diff, corr, iso_inter = map(list,zip(*res))
        return corr, iso_inter



    def get_clusters_features(self):
        psd = self._cluster_prototype_standard_variance()
        cas = self._cluster_average_similarity()
        #noo = self.cluster_number_of_objects()

        self.feature_store = np.array([psd, cas])

        return self.feature_store


    def get_switching_features(self):
        #features list correspond to gene id
        ##iso switch ratio
        ##
        fea1 = self._iso_pairs_switch_ratio()

        self.feature_store = np.array(list(fea1))
        return self.feature_store

    def get_shuffle_switching_features(self):
        fea1 = self._shuffle_iso_pairs_switch_ratio()
        
        self.feature_store = np.array(list(fea1))
        return self.feature_store




    


import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pandas as pd
import scipy.stats
import matplotlib.colors as mcolors
import random

from ..inputdata.DataSet import DataSet
from ..inputdata.BioNetwork import BioNetwork
#from stat.pq_val import _cal_pqvals

def _plot_mean_and_CI(mean, lb, ub, tps, clus, color_mean=None, color_shading=None,  xlab="Time points", ylab="Gene expression"):
    # plot the shaded range of the confidence intervals
    plt.fill_between(range(mean.shape[0]), ub, lb,
                    color=color_shading, alpha=.5)
    # plot the mean on top
    plt.plot(mean, color_mean)
    plt.title('Mean 95% confidence interval of Cluster {}'.format(clus))
    #plt.ylabel("Hyper-similarity score")
    plt.ylabel(ylab)
    plt.xlabel(xlab)
    plt.xticks(np.arange(tps),np.arange(start=1,stop=tps+1,step=1))

def _mean_confidence_interval(data, confidence):
    a = np.array(data)
    n = len(a)
    m, se = np.mean(a), scipy.stats.sem(a)
    h = se * scipy.stats.t.ppf((1 + confidence) / 2., n-1)
    return m, m-h, m+h

def plotting_clusters(clusterObj, confidence=0.95, xlab="Time points", ylab="Gene expression", plot="none"):
    all_g_names = {}
    wilc = []
    t1 = clusterObj.DataSet.timeserieslist[0]
    l = clusterObj.DataSet.l
    index_cluster = clusterObj.index_clusters[0]
    symbs1 = clusterObj.DataSet.symbs[0]
    gene_list = clusterObj.DataSet.gene_list[0]
    tps = clusterObj.DataSet.timepts

    
    col = list(mcolors.CSS4_COLORS.values())

    for u,v in index_cluster.items():
        if len(v) != 0:#len(nms):
            tmm = {}
            for rind in v:
                tmm[str(gene_list[rind])] = symbs1[rind]
            all_g_names[str(u)] = tmm

            ub1 = []
            lb1 = []
            m1 = []
            for ar in t1[list(v)].T:
                x,y,z = _mean_confidence_interval(ar, confidence=0.95)
                m1.append(x)
                lb1.append(y)
                ub1.append(z)
            m1 = np.array(m1)
            ub1 = np.array(ub1)
            lb1 = np.array(lb1)

            wilc.append([t1[list(v)]])
            # fig = plt.figure(1, figsize=(12, 10))
            # plot_mean_and_CI_ylim(m1, ub1, lb1, color_mean='red', color_shading='lightcoral')
            # plt.savefig(other_path + "/" + str(j) +"_"+ args.name + "_red_{}_ylim_conf.pdf".format(j)) 
            # plt.close()
            
            if plot != "all":
                if type(plot) !="list":
                    plot = [plot]
                if u in plot:
                    fig = plt.figure(1, figsize=(12, 10))
                    figshow = _plot_mean_and_CI(m1, ub1, lb1, tps, clus=u, color_mean=random.choice(col), color_shading='lightgrey', xlab=xlab, ylab=ylab)
                    #plt.savefig(other_path + "/" + str(j) +"_"+ args.name + "_red_{}_conf.png".format(j)) 
                    #plt.savefig(other_path + "/" + str(j) +"_"+ args.name + "_red_{}_conf.pdf".format(j)) 
                    print(figshow)
                else:
                    continue

    
    
                # #plt.scatter(d1[:,0], d1[:,1], c="gray")
                # col = ["red", "blue", "green", "yellow", "purple", "white", "brown", "orange", "teal", "silver", "pink"]
                # j = -1
                # for ii, i in use_clu.items():
                #     if len(i) != 0:#len(nms):
                #         #if list(labls).index("3040") in i:
                #         #    print("XXXXXXXXXXXXXXXXXXX", i)
                #         if len(i) >= l: 
                #             j += 1

                #             tmm = {}
                #             for rind in i:
                #                 tmm[str(gene_list[rind])] = symbs1[rind]
                #             all_g_names[str(ii)] = tmm

                #             ub1 = []
                #             lb1 = []
                #             m1 = []
                #             for ar in t1[list(i)].T:
                #                 x,y,z = _mean_confidence_interval(ar, confidence=0.95)
                #                 m1.append(x)
                #                 lb1.append(y)
                #                 ub1.append(z)
                #             m1 = np.array(m1)
                #             ub1 = np.array(ub1)
                #             lb1 = np.array(lb1)

                #             ub2 = []
                #             lb2 = []
                #             m2 = []

                #             for ar in t2[list(i)].T:
                #                 x,y,z = _mean_confidence_interval(ar, confidence=0.95)
                #                 m2.append(x)
                #                 lb2.append(y)
                #                 ub2.append(z)
                #             m2 = np.array(m2)
                #             ub2 = np.array(ub2)
                #             lb2 = np.array(lb2)

                #             wilc.append([t1[list(i)], t2[list(i)]]) #for calculating wilcoxonstats


                #             #print([M1[i].flatten(), M2[i].flatten()])
                            
                #             if plot != "none":
                #                 if type(plot)!="list":
                #                     plot = [plot]
                #                 if ii in plot:
                #                     fig = plt.figure(1, figsize=(12, 10))
                #                     _plot_mean_and_CI(m1, ub1, lb1, tps, clus=ii, color_mean='red', color_shading='lightcoral', xlab=xlab, ylab=ylab)
                #                     _plot_mean_and_CI(m2, ub2, lb2, tps, clus = ii, color_mean='green', color_shading='lightgreen', xlab=xlab, ylab=ylab)
                #                     #plt.savefig(other_path + "/" + str(j) +"_"+ args.name + "_red_green_{}_conf.pdf".format(j)) ## png
                #                     #plt.savefig(other_path + "/" + str(j) +"_"+ args.name + "_red_green_{}_conf.png".format(j)) ## png
                #                     plt.show()
                #                 else:
                #                     continue


    
    clusterObj.DataSet.add_for_graph(all_g_names, wilc)


from setuptools import Extension
from Cython.Build import cythonize
from Cython.Distutils import build_ext
import os


###sklearn cythonize
def configuration(parent_package="", top_path=None):
    from numpy.distutils.misc_util import Configuration, get_numpy_include_dirs
    import numpy
    config = Configuration("_clustering", parent_package, top_path)

    config.add_extension(
        "similarities",
        sources = ['similarities.pyx'],
        include_dirs = [numpy.get_include()]
    )

    return config

if __name__ == "__main__":
    from numpy.distutils.core import setup

    setup(**configuration(top_path="").todict())





import pandas as pd
import numpy as np
import random
import warnings
import math
import itertools as it

from ..DataSet import DataSet
from ..BioNetwork import BioNetwork



class clusterObj():
    def __init__(self,
                genelist_clusters=None,
                index_clusters = None,
                symbs_clusters = None,
                clusters_sim = None,
                _prototype = None,
                _optimal_cluster = None,
                _final_n_cluster = None):
        self.genelist_clusters = genelist_clusters
        self.symbs_clusters = symbs_clusters
        self.index_clusters = index_clusters
        self.clusters_sim = clusters_sim
        self._prototype = _prototype
        self._optimal_cluster = _optimal_cluster
        self._final_n_cluster = _final_n_cluster

    def _get_indexobjects_in_cluster(self, cl):
        indexclu = self.index_clusters
        objs = indexclu[cl]
        objs = np.array(objs, dtype="object")
        return objs

    def _get_idobjects_in_cluster(self, cl):
        idclu = self.genelist_clusters
        objs = idclu[cl]
        objs = np.array(objs, dtype="object")
        return objs

    def _get_symobjects_in_cluster(self, cl):
        symbsclu = self.symbs_clusters
        objs = symbsclu[cl]
        objs = np.array(objs, dtype="object")
        return objs

    



        


    


from .. import constants as constants
import os
g2e_dict = None
e2g_dict = None
dict_type=""

def load_gene_dictionary(gene_list_file_name, gene_list_path=None, source="GDC-TCGA",dataset="melanoma"): #  ="TCGA-SKCM.htseq_counts.tsv"
    if gene_list_path is None:
        gene_list_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "../data",gene_list_file_name)
    f = open(gene_list_path,'r')
    lines = [l.strip() for l in f]
    f.close()
    return lines


def get_g2e_dictionary(cur_dict_type='ENSG'):

    if cur_dict_type=='ENSMUSG':
        lines_dict = load_gene_dictionary(constants.ENSMUSG_TO_GENE_SYMBOLS)
    elif cur_dict_type=='ENSG':
        lines_dict = load_gene_dictionary(constants.ENSG_TO_GENE_SYMBOLS)
    else:
        raise "unknown gene identifiers: {}".format(cur_dict_type)

    gene_symbols2ensembl = {}
    for cur in lines_dict:
        splited_line = cur.split()
        if splited_line[0].find('.') > 0:
            limit = splited_line[0].find('.')
        else:
            limit = len(splited_line[0])
        gene_symbols2ensembl[splited_line[1]] = splited_line[0][:limit]
    return gene_symbols2ensembl

def get_e2g_dictionary(cur_dict_type):
    if cur_dict_type=='ENSMUSG':
        lines_dict = load_gene_dictionary(constants.ENSMUSG_TO_GENE_SYMBOLS)
    elif cur_dict_type=='ENSG':
        lines_dict = load_gene_dictionary(constants.ENSG_TO_GENE_SYMBOLS)
    else:
        raise "unknown gene identifiers: {}".format(cur_dict_type)

    ensembl2gene_symbols = {}
    for cur in lines_dict:
        splited_line = cur.split()
        if splited_line[0].find('.') > 0:
            limit = splited_line[0].find('.')
        else:
            limit = len(splited_line[0])
        ensembl2gene_symbols[splited_line[0][:limit]] = splited_line[1]

    global dict_type
    dict_type=cur_dict_type
    return ensembl2gene_symbols


def e2g_convertor(e_ids):
    if type(e_ids) is str:
        e_ids=[e_ids]

    global g2e_dict
    global dict_type
    cur_dict_type=e_ids[0][:e_ids[0].index('0')]
    if g2e_dict is None or dict_type!=cur_dict_type:
        e2g_dict = get_e2g_dictionary(cur_dict_type)

    results = []
    for cur in e_ids:
        if cur.split(".")[0] in e2g_dict:
            results.append(e2g_dict[cur.split(".")[0]])
        else:
            results.append(cur.split(".")[0])
    return results



import pandas as pd
import numpy as np
import random
import warnings
import logging
import itertools as it

from ..DataSet import DataSet
from ..BioNetwork import BioNetwork
#from . import similarities as sim

#from similarities import calculate_similarities_timeseriesobj_to_medoid, calculate_center_medoid

#from similarity.calculate_similarities import calculate_similarities_timeseriesobj_to_medoid, calculate_similarities_clusterobject, calculate_center_medoid

# def _add_obj_to_most_sim_cluster(timeseriesobjs, nonmedoids_ids, cluster_medoids, cluster_keys, simfunc):
#         ##assign objects to the nearest medoids 
#         ##get clustering
#         ##calculate new medoids
#         clusters={}
        
#         #to cython function
#         #get index for the closest medoid for each object

#         clusters, thisSim = sim.fast_calculate_similarities_timeseriesobj_to_medoid(timeseriesobjs, nonmedoids_ids, cluster_medoids, simfunc)
        
#         return clusters, thisSim

# def _new_add_obj_to_most_sim_cluster(nonsampledata, nonsampledata_id, mapped_clusters, prototypes, simfunc):
#     #calculate similarities within clusters
#     #get clusters number for the closest

#     if simfunc == "pearson":
#         mat = [sim.corr_mat(nonsampledata[x], prototypes) for x in range(nonsampledata.shape[0])]
        
#     if simfunc == "euclidean":
#         mat = [sim.dist_mat(nonsampledata[x], prototypes) for x in range(nonsampledata.shape[0])]

#     mat = np.average(mat, axis=0)
#     #maxforobj is the index of medoid which the timeseries is closest
#     maxforobj = np.argmax(mat, axis=1)

#     for id, maxi in enumerate(maxforobj):
#         if str(maxi+1) in mapped_clusters.keys():
#             mapped_clusters[str(maxi+1)].append(nonsampledata_id[id])
#         else:
#             mapped_clusters.setdefault(str(maxi+1), [nonsampledata_id[id]])
    

#     return mapped_clusters

# def within_cluster_similarity(alltimeseries, clusters, similarityfunction):
#     results = np.zeros(len(clusters), dtype="double")
#     c=0
#     for u,v in clusters.items():
#         if similarityfunction == "pearson":
#             mat = [sim.corr_mat(alltimeseries[x,v,:], alltimeseries[x,v,:]) for x in range(alltimeseries.shape[0])]

#         if similarityfunction == "euclidean":
#             mat = [sim.dist_mat(alltimeseries[x,v,:], alltimeseries[x,v,:]) for x in range(alltimeseries.shape[0])]
#         #take average and along axis 1
#         matq = np.average(mat, axis=0)
#         within_sim = np.sum(matq, axis=1)
#         results[c] = within_sim[np.argmax(within_sim)]/len(v)
#         c+=1  

#     return results

def cluster_map_to_item(DataSet, clusters):
    gene_list = DataSet.gene_list
    symbs = DataSet.symbs
    
    g_clusters = {}
    s_clusters = {}


    for u,v in clusters.items():
        g_tmp = [np.array(gene_list)[x] for x in v]
        s_tmp = [np.array(symbs)[x] for x in v]

        g_clusters["Cluster "+str(u)] = g_tmp
        s_clusters["Cluster "+str(u)] = s_tmp

    return g_clusters, s_clusters

def merge_above_similarity_threshold_clusters(clusters, prototypes, threshold=0.9):
    sims = []
    for k in it.combinations(list(prototypes.keys()),2):
        tmp_sims = sim.pearson_corr(np.array(prototypes[k[0]][0]), np.array(prototypes[k[1]][0]), len(prototypes[k[0]]))
        print(k)
        print(tmp_sims)
        # if tmp_sims > threshold:
        #     #make clusters prototypes a dict
        #     clusters[k[0]].append(clusters[k[1]])
        #     clusters.pop(k[1])
        #     prototypes.pop(k[1])

    return clusters, prototypes

    


# this script is derived from sklearn-extra packages KMedoids

import numpy as np
import warnings

from sklearn.base import BaseEstimator, ClusterMixin, TransformerMixin

from . import similarities as sim

class KMedoids(BaseEstimator, ClusterMixin, TransformerMixin):
    def __init__(
        self,
        n_clusters=8,
        metric="precomputed",
        max_iter=300,
        random_state=None,
        method="pam"
    ):
        self.n_clusters = n_clusters
        self.metric = metric
        self.max_iter = max_iter
        self.random_state = random_state
        self.method = method

    def _initialize_medoids(self, D, n_clusters):
        """define initial medoids:
        heuristic""" 

        medoids = np.random.choice(D.shape[0], n_clusters)
        return medoids


    def _compute_inertia(self, distances):
        """Compute inertia of new samples. Inertia is defined as the sum of the
        sample distances to closest cluster centers.
        Parameters
        ----------
        distances : {array-like, sparse matrix}, shape=(n_samples, n_clusters)
            Distances to cluster centers.
        Returns
        -------
        Sum of sample distances to closest cluster centers.
        """

        # Define inertia as the sum of the sample-distances
        # to closest cluster centers
        inertia = np.sum(np.min(distances, axis=1))

        return inertia

    def transform(self, X):
        """Transforms X to cluster-distance space.
        Parameters
        ----------clu = clustering(allcovdata, algorithm='kmedoids', metric='correlation',BioNetwork=bionet, composite=False)
c = clu.find_clusters()

%matplotlib inline
cluster_heatmap(clu, standard_scale=0)
            X transformed in the new space of distances to cluster centers.
        """
    
        if self.metric == "precomputed":
            return X[:, self.medoid_indices_]

    def _update_medoid_idxs_in_place(self, D, labels, medoid_idxs):
        """In-place update of the medoid indices"""

        # Update the medoids for each cluster
        for k in range(self.n_clusters):
            # Extract the distance matrix between the data points
            # inside the cluster k
            cluster_k_idxs = np.where(labels == k)[0]

            if len(cluster_k_idxs) == 0:
                warnings.warn(
                    "Cluster {k} is empty! "
                    "self.labels_[self.medoid_indices_[{k}]] "
                    "may not be labeled with "
                    "its corresponding cluster ({k}).".format(k=k)
                )
                continue

            in_cluster_distances = D[
                cluster_k_idxs, cluster_k_idxs[:, np.newaxis]
            ]

            # Calculate all costs from each point to all others in the cluster
            in_cluster_all_costs = np.sum(in_cluster_distances, axis=1)

            min_cost_idx = np.argmin(in_cluster_all_costs)
            min_cost = in_cluster_all_costs[min_cost_idx]
            curr_cost = in_cluster_all_costs[
                np.argmax(cluster_k_idxs == medoid_idxs[k])
            ]

            # Adopt a new medoid if its distance is smaller then the current
            if min_cost < curr_cost:
                medoid_idxs[k] = cluster_k_idxs[min_cost_idx]

    def _compute_cost(self, D, medoid_idxs):
        """ Compute the cose for a given configuration of the medoids"""
        return self._compute_inertia(D[:, medoid_idxs])

    def fit(self, X):
        """
        x: distance array
        """
        if self.n_clusters > X.shape[0]:
            raise ValueError(
                "The number of medoids (%d) must be less "
                "than the number of samples %d."
                % (self.n_clusters, X.shape[0])
            )

        medoid_idxs = self._initialize_medoids(X, self.n_clusters)

        #compute the distance to the first and second closest points among medoids
        Djs, Ejs = np.sort(X[medoid_idxs], axis=0)[[0,1]]


        for self.n_iter_ in range(0, self.max_iter):
            old_medoid_idxs = np.copy(medoid_idxs)
            labels = np.argmin(X[medoid_idxs,:], axis=0)

            if self.method == "alternate":
                # Update medoids with the new cluster indices
                self._update_medoid_idxs_in_place(D, labels, medoid_idxs)
            
            elif self.method == "pam":
                not_medoid_idxs = np.delete(np.arange(len(X)), medoid_idxs)
                optimal_swap = sim._compute_optimal_swap(X,
                                                        medoid_idxs.astype(np.intc),
                                                        not_medoid_idxs.astype(np.intc),
                                                        Djs,
                                                        Ejs,
                                                        self.n_clusters)

                if optimal_swap is not None:
                    i,j, _ = optimal_swap
                    medoid_idxs[medoid_idxs == i] = j

                    #update
                    Djs, Ejs = np.sort(X[medoid_idxs], axis=0)[[0,1]]

            if np.all(old_medoid_idxs == medoid_idxs):
                break
            elif self.n_iter_ == self.max_iter - 1:
                warnings.warn(
                    "Maximum number of iteration reached before "
                    "convergence. Consider increasing max_iter to "
                    "improve the fit."
                )

            # Expose labels_ which are the assignments of
            # the training data to clusters
            self.labels_ = np.argmin(X[medoid_idxs, :], axis=0)
            self.medoid_indices_ = medoid_idxs
            self.inertia_ = self._compute_inertia(self.transform(X))

            # Return self to enable method chaining
            return self


import pandas as pd
import numpy as np
import time

from ..DataSet import DataSet
from ..BioNetwork import BioNetwork

def _prototype_builder(timeseriesobj, clusters, prototype_func):
    '''
    ##calculate prototype of clusters
    #medoid
    ##median or average

    return dict:
    {'34': array([-0.14823925, -0.74057195,  0.41986664, -0.24045551]),
    '146': array([-0.20850356,  0.29922031,  0.74922338,  0.31151074]),
    '195': array([ 0.26338935,  0.19041125,  0.22270178, -0.19392188]),
    '27': array([-0.27389526, -0.6797392 ,  0.98956317,  0.29812246]),
    '78': array([ 0.19380661, -0.91636647,  0.44639439, -0.18150735]),
    '168': array([-1.6519947 , -0.25548007, -0.3339432 , -0.28108202])}
    '''
    cluster_pro = {}
    len_ts = timeseriesobj.shape[0]
    mean_items = 0

    c=0
    for m, n in clusters.items():
        timeseriesobjdf_list = []
        if len(n)>=1:
            timeseriesobj_list = np.asarray(timeseriesobj)[:,list(n),:]
            
            if prototype_func == "mean":
                try:
                    mean_items = np.mean(np.mean(timeseriesobj_list, axis=0), axis=0)
                except:
                    mean_items = np.mean(timeseriesobj_list, axis=0)
            if prototype_func == "median":
                try:
                    mean_items = np.median(np.median(timeseriesobj_list, axis=0), axis=0)
                except:
                    mean_items = np.median(timeseriesobj_list, axis=0)
        
            cluster_pro.setdefault(m, np.array(mean_items, dtype="double"))
        
        
        else:
            continue
        c+=1
    
    for i in clusters.keys():
        if i not in cluster_pro.keys():
            print(i)

    return cluster_pro


from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name='domino_hagai',
    version="0.1",
    author="Hagai Levi",
    author_email="hagai.levi.007@gmail.com",
    description='DOMINO: Discovery of Modules In Networks using Omic',
    url='https://github.com/Shamir-Lab/DOMINO',
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Linux",
    ],
    packages = find_packages(),
    package_data={'': ['*']},
    include_package_data=True,
    install_requires=[
        'networkx==2.4',
        'matplotlib==3.1.3',
        'numpy==1.18.1',
        'scipy==1.4.1',
        'pandas==1.0.1',
        'pcst-fast==1.0.7',
        'statsmodels==0.11.0',
        'python-louvain==0.14'],
    entry_points = {
        "console_scripts": [
            "domino=src.runner:main_domino",
            "slicer=src.runner:main_slicer",
        ]
    }

)




##network class
#import networkx as nx
import numpy as np
import networkx as nx
import os
dir_path = os.path.dirname(os.path.realpath(__file__))

class BioNetwork:
    """Storage of biological network.

    Parameters
    ----------
    path : 
        dir path to the network file or "human" as default human biogrid network


    Attributes
    -----------
    connected_nodes 

    total_node_degree_counts

    total_undirected_edgecount_nodedegree

    max_degree

    all_degree

    Methods
    ----------
    g : 
        generate network from file path if type is `file` or return itself if type is `network`

    lst_g : 
        return list of node names

    adj : 
        return adjcency matrix

    removing_nodes : 
        remove nodes given the index


    """
    def __init__(
        self, 
        path = "human",
        keytype="entrezid",
        **kwargs
        ):
        self.path = path
        self.keytype=keytype

        self.kwargs = {}
        self.kwargs.update((key, value) for key, value in kwargs.items())
        # self.connected_nodes = connected_nodes
        # self.total_node_degree_counts = total_node_degree_counts
        # self.total_undirected_edgecount_nodedegree = total_undirected_edgecount_nodedegree
        # self.max_degree = max_degree
        # self.all_degree = all_degree
        
   

        
        """
        Read the network path and generate networkx object.
        """
        if self.path == "human":
            #print("Reading edge list and removing loops")
            path = os.path.join(dir_path,f'data/network/9606_biogrid_entrez.tab')
            g = nx.read_edgelist(path=path, **self.kwargs)
            tmp = 0
            for gene in g:
                if gene in g[gene]:
                    tmp += 1
                    g.remove_edge(gene, gene)
        elif self.path == "mouse":
            path = os.path.join(dir_path,f'data/network/10090_biogrid_entrez.tab')
            g = nx.read_edgelist(path=path, **self.kwargs)
            tmp = 0
            for gene in g:
                if gene in g[gene]:
                    tmp += 1
                    g.remove_edge(gene, gene)
        else:
            g = nx.read_edgelist(path=path, **self.kwargs)
            tmp = 0
            for gene in g:
                if gene in g[gene]:
                    tmp += 1
                    g.remove_edge(gene, gene)
            # #print("Read edge list, network contains {} nodes and {} edges".format(len(g.nodes()), len(g.edges())))
            # #print('Removed', tmp, 'loops')

            #use graph-tool
            # edgelist = []
            # with open(self.path, "r") as f:
            #     for a in f.readlines():
            #         edgelist.append((a[:-1].split("\t")[0],a[:-1].split("\t")[1]))

            # g = Graph(directed=False)

            # props = g.add_edge_list(edgelist, hashed=True)
            # v_prop = g.new_vertex_property("string")
            # for i in range(g.num_vertices()):
            #     v_prop[i] = props[i]

            # g.vertex_properties['name'] = v_prop
            # remove_parallel_edges(g) #remove duplicates edges
        self.networkz = g

    def g(self):
        """
        Return the networkx object.
        """
        return self.networkz
        
    def lst_g(self):
        """
        Return list of gene name in the network.
        """
        g = self.networkz
        lst_g = list(g)
        lst_g.sort()

        # for i in range(g.num_vertices()):
        #     lst_g.append(g.vertex_properties['name'][i])
        #lst_g = [str(x) for x in lst_g]
        return np.array(lst_g)

    def adj(self):
        """
        Return adjacency matrix of the network.
        """
        #print("computing adjacency matrix...")
        g = self.networkz
        lst_g = self.lst_g()
        adj = nx.to_numpy_matrix(g, nodelist=lst_g)
        
        #graph tool
        # g = self.g()
        # lst_g = self.lst_g()
        # a = adjacency(g)
        # adj = a.A #csr matrix to array
        # adj[adj>1] = 1 
        # return np.array(adj)
        return np.array(adj)


    # def read_graph(self):
    #     self.g = self._getgraph(self.path)
    #     # self.lst_g = self._getgene(self.g)
    #     # self.adj = self._getadj(self.g, self.lst_g)

    def _removing_nodes(self, nodes_to_remove):
        #nodes_to_remove = nodes name
        network = self.networkz

        #network.remove_vertex(nodes_to_remove) #graphtool
        
        network.remove_nodes_from(nodes_to_remove)
        self.networkz = network


    def _get_connected_nodearray(self):
        # if self.connected_nodes is None:
        #     #basically edge list
        #     network = self.g()
        #     connected_nodes = list(network.get_edges())

        #     self.connected_nodes = connected_nodes
        #     return connected_nodes
        # else:
        #     return self.connected_nodes
        if "self.connected_nodes" not in locals():
            #basically edge list
            network = self.networkz
            connected_nodes = list(network.edges)

            self.connected_nodes = connected_nodes
            return connected_nodes
        else:
            return self.connected_nodes


    def _get_total_node_degree_counts(self):
        # if self.total_node_degree_counts is None:
        #     #return counts of node degree for all nodes
        #     network = self.g()
        #     #return array
        #     nodedegrees = network.degree_property_map("total").get_array()
        #     result = np.empty((max(nodedegrees)+1), dtype="double")
        #     result.fill(0)
        #     for i in nodedegrees:
        #         result[i] += 1
            
        #     self.total_node_degree_counts = result
        #     return result
        # else:
        #     return self.total_node_degree_counts
        if "self.total_node_degree_counts" not in locals():
            #return counts of node degree for all nodes
            network = self.networkz
            nodedegrees = dict(network.degree())
            result = np.empty((max(nodedegrees.values())+1), dtype="double")
            result.fill(0)
            for i in nodedegrees.values():
                result[i] += 1
            
            self.total_node_degree_counts = result
            return result
        else:
            return self.total_node_degree_counts

        

    def _get_undirected_node_degree(self) -> dict:
        network = self.networkz
        #nodedegrees = network.degree_property_map("total").get_array()
        nodedegrees = dict(network.degree())

        self.all_degree = nodedegrees
        return self.all_degree
        
        

    def _get_maxdegree(self):
        #return the max node degree
        if "self.max_degree" not in locals():
            network = self.networkz
            # nodedegrees = network.degree_property_map("total").get_array()
            # maxdegree = max(nodedegrees)
            nodedegrees = dict(network.degree())
            maxdegree = max(nodedegrees.values())

            self.max_degree = maxdegree

            return self.max_degree
        else:
            return self.max_degree

    def _get_total_undirected_edge_count_nodedegrees(self):
        if "self.total_undirected_edgecount_nodedegree" not in locals():
            #return total node degree for edges for undirected graph
            nodedegrees = dict(self.networkz.degree())
            maxdegree = self._get_maxdegree() 
            total_undirected_edgecount_nodedegree = np.empty((maxdegree+1, maxdegree+1), dtype="double")
            total_undirected_edgecount_nodedegree.fill(0)
            connectednodes = self._get_connected_nodearray()
            
            for i in connectednodes:
                node1 = i[0]
                node2 = i[1]
                degree1 = nodedegrees[node1]
                degree2 = nodedegrees[node2]

                mindeg = min(degree1, degree2)
                maxdeg = max(degree1, degree2)
                
                total_undirected_edgecount_nodedegree[mindeg][maxdeg] += 1

            self.total_undirected_edgecount_nodedegree = total_undirected_edgecount_nodedegree
            return total_undirected_edgecount_nodedegree
        else: 
            return self.total_undirected_edgecount_nodedegree



    def _get_total_directed_edge_count_nodedegrees(self):
        try:
            total_directed_edgecount_nodedegree = self.total_directed_edgecount_nodedegree
        
        except:
            pass
        
        return total_directed_edgecount_nodedegree


import numpy as np
import pandas as pd
from collections import defaultdict
from scipy.stats import mannwhitneyu
from scipy.stats import chi2
import warnings
from functools import reduce
from itertools import combinations
import pickle, os, sys
warnings.simplefilter("ignore")
from collections import defaultdict

from ._util_stat.multipletesting import pvalue_correction
from ._feature import featuresObj 
from ._shuffle import shuffling
# Disable
def _blockPrint():
    sys.stdout = open(os.devnull, 'w')

# Restore
def _enablePrint():
    sys.stdout = sys.__stdout__

dir_path = os.path.dirname(os.path.realpath(__file__))

# def _iso_switch_between_arrays(arr1, arr2, orgarr1, orgarr2):
#         ab = np.sign(arr1-arr2)

#         ##look for switch points
#         points=[[x for x in range(ab.shape[1]) if ab[r,x] != np.roll(ab[r], 1)[x] and x!=0] for r in range(ab.shape[0])]

#         #points: switch points of all replicates
#         ##check if all r have switch points
#         #then take the points where all appeared
#         def multiple_intersect(points):
#             return np.unique(list(reduce(lambda x, y: set(x).intersection(set(y)), points)))

#         #calculate differences
#         #[(abs(a[x-1] - a[x]) + abs(b[x-1] -b[x]))/2 for x in points], a,b
#         if any(points):
#             final_sp = multiple_intersect(points) 
#             if not any(final_sp):
#                 final_sp = [list(set(x[0]).intersection(set(x[1]))) for x in combinations(points,2)]
#                 final_sp = np.unique(reduce(lambda x,y: x+y, final_sp))
#                 if not any(final_sp):
#                     final_sp = reduce(lambda x,y : x+y, points)
        

#             final_sp = np.sort(final_sp)
#             # mean of max difference at the time points between 2 arrays for all replicates
#             allsp_diff = [np.mean([np.max([abs(arr1[r,x-1] - arr2[r,x-1]), abs(arr2[r,x] -arr1[r,x])]) for r in range(arr1.shape[0])]) for x in final_sp]
            
#         else:
#             final_sp = []
#             allsp_diff=[]

#         #calculate corr
#         cors = [(1-np.corrcoef(orgarr1[r], orgarr2[r])[0,1])/2 for r in range(arr1.shape[0])]
        
#         #for each switch point, the max switch diff before and after 
#         #allsp : all the before and after switch differences for all switch points
#         return len(final_sp)/arr1.shape[1], allsp_diff, final_sp, np.mean(cors)
class basic_pvalue():
    def __init__(self, testobject, object_type, iso_switch = True, shuffle=None, fitness_scores = None, fitness_scores_two_sided = False, n_permutations=1000):
        self.testobject=testobject ##DataSet
        self.object_type = object_type #connectivity, clusters, iso_pairs
        self.shuffle = shuffle
        self.fitness_scores = fitness_scores
        self.fitness_scores_two_sided = fitness_scores_two_sided
        #self.combine_pvalues = combine_pvalues
        self.n_permutations = n_permutations
        self.iso_switch = iso_switch

    def initialize_features_for_permutation(self, features, shuffled):
        pass

    def calculate_original_object_fitness(self):
        pass

    # Print iterations progress
    def progressbar(self, it, prefix="", size=60, file=sys.stdout):
        count = len(it)
        def show(j):
            x = int(size*j/count)
            file.write("%s[%s%s] %i/%i\r" % (prefix, "#"*x, "."*int((size-x)), j, count))
            file.flush()        
        show(0)
        for i, item in enumerate(it):
            yield item
            show(i+1)
        file.write("\n")
        file.flush()


    def empirical_pvalue(self, total_observation,better_observation):
        if total_observation > 0:
            result = (1+better_observation)/ (1+total_observation)
            return result
        
    def combine_pvals(self, pvals):
        chisqprob = lambda chisq, df: chi2.sf(chisq, df)
        s = -2 * np.sum(np.log(pvals))
        
        return chisqprob(s, 2 * len(pvals))

        
    def do_calculate(self):
        if self.object_type=="clusters":
            n_object = self.testobject.clusterobj._final_n_cluster
            timeserieslist = self.testobject.timeserieslist
            #get features of clusters
            print("Calculating features.")
            features_original = featuresObj(self.testobject, timeserieslist, feature_type="clusters")
            #get fitness of clusters
            print("Calculating fitness score.")
            original_fitness_score = features_original.feature_store
        

            #data structure for storing permuted features #[features][permutation]
            permuted_fitness = np.zeros((original_fitness_score.shape[0], n_object))

        # else:
        #     if not isinstance(testobject, clusterObj):
        #         raise ValueError("ClusterObj object is needed.")
            
        if self.object_type == "iso_pairs":
            n_object = len(self.testobject.isoobj.isopairs['major_transcript'])
            timeserieslist = self.testobject.timeserieslist
            features_original = featuresObj(self.testobject, timeserieslist, feature_type="iso_pairs")
            original_fitness_score = features_original.feature_store
            permuted_fitness = np.zeros((original_fitness_score.shape[0], n_object))
 

        batch_smaller=np.zeros((original_fitness_score.shape[0], n_object))
        batch_larger=np.zeros((original_fitness_score.shape[0], n_object))
        batch_equal=np.zeros((original_fitness_score.shape[0], n_object))
        batch_comp=np.zeros((original_fitness_score.shape[0], n_object))

        #shuffle feature and fitness  

        #def _cal_permutations(p):
        for p in self.progressbar(it=range(self.n_permutations), prefix="Permutation:", size=self.n_permutations/10):
            _blockPrint()
  
            shuffleObj = shuffling(timeserieslist=timeserieslist)
            #shuffle1 = np.array(shuffleObj.shuffle_dataset_rowwise(), dtype="double")
            #self.testobject.timeserieslist = shuffle1 ##reassign testobject time series to the shuffled dataset
            
            #calculate features for shuffle obj
            iso = iso_function(self.testobject)
            #corrdf = iso.detect_isoform_switch(filtering=False, min_diff=self.testobject.isoobj.min_diff, corr_cutoff=self.testobject.isoobj.corr_cutoff, event_im_cutoff=self.testobject.isoobj.event_im_cutoff)
            #self.testobject = iso.total_isoform_usage(corrdf)
            #shuffle_clu = clustering.clustering(self.testobject, input_type="isoformusage", algorithm=self.testobject.clusterobj.algorithm, metric=self.testobject.clusterobj.metric, linkage=self.testobject.clusterobj.linkage, n_clusters=self.testobject.clusterobj.n_clusters)


            if self.object_type=="clusters":
                permuted_features = featuresObj(testobject=self.testobject, timeserieslist=timeserieslist, feature_type = "clusters", seed=p)
            #calculate fitness for shuffle obj 
                permuted_fitness = permuted_features.feature_store


            elif self.object_type=="iso_pairs":
                permuted_features = featuresObj(testobject=self.testobject, timeserieslist = timeserieslist, feature_type="shuffle_iso_pairs", seed=p)
                permuted_fitness = permuted_features.feature_store

            
            for fs in range(original_fitness_score.shape[0]):
                fitness = original_fitness_score[fs]
                shuffledfitness = permuted_fitness[fs]
                    
                #comparison
                for o in range(len(fitness)):
                    compare = shuffledfitness[o] - fitness[o]

                    if compare < 0:
                        batch_smaller[fs][o] += 1
                    elif compare > 0:
                        batch_larger[fs][o] += 1
                    else:
                        batch_equal[fs][o] += 1
                    batch_comp[fs][o] += 1

            p+=1
            _enablePrint()
           

        #res = Parallel(backend="multiprocessing") (delayed(_cal_permutations)(int(p)) for p in self.progressbar(it=range(self.n_permutations), prefix="Permutation:", size=self.n_permutations/10))
        #print(res)
       
        #end permutation
        print("Done permutation")
        pvals=np.zeros((original_fitness_score.shape[0], n_object))
        total_observation = 0
        better_observation = 0
        

        for fs in range(original_fitness_score.shape[0]):
            for o in range(n_object):
                total_observation = batch_comp[fs][o]
                

                if self.fitness_scores_two_sided is True:
                    better_observation = min(batch_larger[fs][o], batch_smaller[fs][o])*2 + batch_equal[fs][o]
                else:
                    better_observation = batch_equal[fs][o] + batch_larger[fs][o]

                pvals[fs][o] = self.empirical_pvalue(total_observation, better_observation)
                
        # p_vals = Parallel(backend="multiprocessing") (delayed(_count_permutations)(fs, o) for o in range(n_object) for fs in range(original_fitness_score.shape[0])) 
       
        #p_vals = np.array(p_vals).reshape(n_object, original_fitness_score.shape[0])

        ##combine pvalues
        result_pvalue = np.zeros(n_object)

        for o in range(n_object):
            p_values_list = np.zeros(original_fitness_score.shape[0])
            p_values_list = pvals[:,o]
            result_pvalue[o] = self.combine_pvals(p_values_list)

        return result_pvalue

class iso_function():
    """
    Isoform level analysis

    Parameters 
    ----------
    DataSet : 
        DataSet object
    Species : 
        Species ID

    Return 
    --------
    Create an instance for isoform switch analysis. 
    
    """
    def __init__(self, dataset):
        self.dataset = dataset
        self.species = dataset.species
    
        if "." in dataset.transcript_id[0]:
            infile = open(os.path.join(dir_path,f"data/network/{dataset.species}_ver_trans_pfam.pkl"),'rb')
            exfile = open(os.path.join(dir_path,f"data/network/exons_{dataset.species}_ensembl.pkl"),'rb')
        else:
            infile = open(os.path.join(dir_path,f"data/network/{dataset.species}_ver_trans_pfam.pkl"),'rb')
            exfile = open(os.path.join(dir_path,f"data/network/exons_{dataset.species}_ensembl.pkl"),'rb')
        self.trans_pfam_dict = pickle.load(infile)
        self.trans_exon_dict = pickle.load(exfile)

        groupdict = defaultdict(lambda: defaultdict(list))
        for idx, sym in enumerate(self.dataset.gene_id):
            if sym != 'n':
                groupdict[sym]['trans_id'].append(self.dataset.transcript_id[idx])
                groupdict[sym]['indices'].append(idx)
                groupdict[sym]['array'].append(self.dataset.timeserieslist[:,idx,:])
        
        #normlize work
        for sym, df in groupdict.items():
            org = df['array'][0]
            for x in range(1,len(df['array'])):
                org = np.hstack([org, df['array'][x]])
            
            df['array'] = org.reshape(self.dataset.reps1, len(df['array']), self.dataset.timepts)
            
            tmp = np.array([df['array'][x]/np.sum(df['array'][x], axis=0) for x in range(df['array'].shape[0])])
            tmp[np.isnan(tmp)] = 1
            groupdict[sym]['normarr'] = tmp

        self.normdf = groupdict

    def _take_interval(self, sp, timepts):
        newsp = np.insert(sp,0,0)
        newsp = np.append(newsp, timepts)
        intervals = []
        for i in range(len(sp)):
            intervals.append(np.min([newsp[i+1]-newsp[i], newsp[i+2]-newsp[i+1]]))
        return intervals

    def _iso_switch_between_arrays(self, arr1, arr2, orgarr1, orgarr2):
        ab = np.sign(arr1-arr2)

        if self.dataset.reps1 ==1:
            points = [x for x in range(1, ab.shape[1]) if (ab[0][x] != np.roll(ab[0], 1)[x])]
            final_sp=points
            allsp_diff =[np.mean([abs(arr1[:,x-1] - arr2[:,x-1]), abs(arr2[:,x] -arr1[:,x])]) for x in final_sp]
            cors = (1-np.corrcoef(orgarr1, orgarr2)[0,1])/2
            iso_ratio=len(final_sp)/arr1.shape[0]
        else:
        ##look for switch points
            points=[[x for x in range(ab.shape[1]) if ab[r,x] != np.roll(ab[r], 1)[x] and x!=0] for r in range(ab.shape[0])]
        #points: switch points of all replicates
        ##check if all r have switch points
        #then take the points where all appeared
            def multiple_intersect(points):
                return np.unique(list(reduce(lambda x, y: set(x).intersection(set(y)), points)))

            #calculate differences
            #[(abs(a[x-1] - a[x]) + abs(b[x-1] -b[x]))/2 for x in points], a,b
            rep_agree = round(self.dataset.reps1*0.6)
            if any(points):
                final_sp = multiple_intersect(points) 
                if not any(final_sp):
                    final_sp = [list(set(x[0]).intersection(set(x[1]))) for x in combinations(points,rep_agree)]
                    final_sp = np.unique(reduce(lambda x,y: x+y, final_sp))
                    if not any(final_sp):
                        final_sp = []
            

                final_sp = np.sort(final_sp)
                # mean of max difference at the time points between 2 arrays for all replicates
                allsp_diff = [np.mean([np.max([abs(arr1[r,x-1] - arr2[r,x-1]), abs(arr2[r,x] -arr1[r,x])]) for r in range(arr1.shape[0])]) for x in final_sp]
                
            else:
                final_sp = []
                allsp_diff=[]

        
        #calculate corr
            cors = [(1-np.corrcoef(orgarr1[r], orgarr2[r])[0,1])/2 for r in range(arr1.shape[0])]
            iso_ratio = len(final_sp)/arr1.shape[1]
        #for each switch point, the max switch diff before and after 
        #allsp : all the before and after switch differences for all switch points
        return iso_ratio, allsp_diff, final_sp, np.mean(cors)


    def _event_enrichness(self, normdf, sp, arr1, arr2):
        ##penalize low expressed genes
        penal1 = np.mean([arr1[r]/np.max(normdf[r],axis=0) for r in range(normdf.shape[0])],axis=0)
        penal2 = np.mean([arr2[r]/np.max(normdf[r],axis=0) for r in range(normdf.shape[0])],axis=0)

        return np.mean(np.append(penal1[sp-1:sp+1],penal2[sp-1:sp+1]))

        
    def _remove_low_expressed(self):
        #return isoform indices
        timedf = pd.concat([pd.DataFrame(self.timeserieslist[x]) for x in range(self.timeserieslist.shape[0])], axis=1)
        geneexp = pd.concat([pd.Series(self.symbs, name='symbol'), timedf],axis=1)

        #filter out those higher than 2 by default (can change)
        lowgenes_idx = np.where(geneexp.groupby('symbol').agg('sum').mean(axis=1) < 2)[0]
        #get the gene names lower than cutoff
        lowgenes = geneexp.iloc[lowgenes_idx, 0] 


        return lowgenes

    def _remove_low_abundance(self, pre_normdf):
        #return isoform indices
    

        #calculate fold change of major isoform with respect to the minor isoforms
        lowexp = np.mean(np.max(pre_normdf, axis=2),axis=0)
        
        #filter out those higher than 2 by default (can change)
        lowgenes = np.where(lowexp<=0.1)[0]
        
        filtered_normdf = np.delete(pre_normdf, lowgenes, axis=1)
        return filtered_normdf, lowgenes

    def _combine_pvals(self, pvals):
        ##no used
        chisqprob = lambda chisq, df: chi2.sf(chisq, df)
        s = -2 * np.sum(np.log(pvals))
        
        return chisqprob(s, 2 * len(pvals))

        
    def _take_interval(self, sp):
        #check interval between switch points
        newsp = np.insert(sp,0,0)
        newsp = np.append(newsp, self.dataset.timepts)
        intervals = []
        for i in range(len(sp)):
            intervals.append(np.min([newsp[i+1]-newsp[i], newsp[i+2]-newsp[i+1]]))
        return intervals

    def _diff_before_after_switch(self, normdf, arr1, arr2, allsp_diff, switch_points):
        tmp_sw =  np.insert(switch_points,0,0)
        tmp_sw = np.append(tmp_sw, self.dataset.timepts)
        if self.dataset.reps1>1:
            #get the intervals between time points
            if any(switch_points):
                #the best sp has the highest mean differences before and after switch points between the two time series 
                bestpval1=0
                bestpval2=0
                allbestp=1
                finaldiff = 0
                finalenrich = 0
                best_switch_point=0
                this_prob = 0
                
                thispt=0 

                for bp in range(1, len(tmp_sw)-1):
                    #if len(switch_points) != bp+1:
                    iso1_pval = mannwhitneyu(arr1[:,tmp_sw[bp-1]:tmp_sw[bp]].reshape(1,-1)[0], arr1[:,tmp_sw[bp]:tmp_sw[bp+1]].reshape(1,-1)[0])
                    iso2_pval = mannwhitneyu(arr2[:,tmp_sw[bp-1]:tmp_sw[bp]].reshape(1,-1)[0], arr2[:,tmp_sw[bp]:tmp_sw[bp+1]].reshape(1,-1)[0])
                    thisdiff = np.mean([abs(np.mean(arr1[:,tmp_sw[bp-1]:tmp_sw[bp]]) - np.mean(arr1[:,tmp_sw[bp]:tmp_sw[bp+1]])), abs(np.mean(arr2[:,tmp_sw[bp-1]:tmp_sw[bp]]) - np.mean(arr2[:,tmp_sw[bp]:tmp_sw[bp+1]]))])
                    thisenrich = self._event_enrichness(normdf, tmp_sw[bp], arr1, arr2)
                    iso_prob = ((np.sum(arr1[:,tmp_sw[bp-1]:tmp_sw[bp]].reshape(1,-1)[0]>arr2[:,tmp_sw[bp-1]:tmp_sw[bp]].reshape(1,-1)[0]))/arr1[:,tmp_sw[bp-1]:tmp_sw[bp]].reshape(1,-1)[0].shape[0]) + ((np.sum(arr1[:,tmp_sw[bp]:tmp_sw[bp+1]].reshape(1,-1)[0]<arr2[:,tmp_sw[bp]:tmp_sw[bp+1]].reshape(1,-1)[0]))/(arr1[:,tmp_sw[bp]:tmp_sw[bp+1]].reshape(1,-1)[0].shape[0]))

                    thispt=tmp_sw[bp]

                    if (iso1_pval[1] < allbestp) & (thisenrich > finalenrich):
                        bestpval1, bestpval2=iso1_pval[1], iso2_pval[1]
                        allbestp=iso1_pval[1]
                        finaldiff = thisdiff
                        finalenrich = thisenrich
                        best_switch_point=tmp_sw[bp]
                        this_prob = iso_prob/2
                    elif (iso2_pval[1] < allbestp) & (thisenrich > finalenrich):
                        bestpval1, bestpval2=iso1_pval[1], iso2_pval[1]
                        allbestp=iso2_pval[1]
                        finaldiff = thisdiff
                        finalenrich = thisenrich
                        best_switch_point=tmp_sw[bp]
                        this_prob = iso_prob/2
                    else:
                        continue

                return finaldiff, bestpval1, bestpval2, best_switch_point, finalenrich, this_prob
            else:
                return 0, 1, 1, 0, 0, 0

        else:             
            #the best sp has the highest mean differences before and after switch points between the two time series 
            if any(allsp_diff):
                bp = np.argmax(allsp_diff)
                finalenrich = self._event_enrichness(normdf, bp, arr1, arr2)
                iso_prob = ((np.sum(arr1[:,tmp_sw[bp-1]:tmp_sw[bp]].reshape(1,-1)[0]>arr2[:,tmp_sw[bp-1]:tmp_sw[bp]].reshape(1,-1)[0]))/arr1[:,tmp_sw[bp-1]:tmp_sw[bp]].reshape(1,-1)[0].shape[0]) + ((np.sum(arr1[:,tmp_sw[bp]:tmp_sw[bp+1]].reshape(1,-1)[0]<arr2[:,tmp_sw[bp]:tmp_sw[bp+1]].reshape(1,-1)[0]))/(arr1[:,tmp_sw[bp]:tmp_sw[bp+1]].reshape(1,-1)[0].shape[0]))
                return allsp_diff[bp], 1, 1, switch_points[bp], finalenrich, iso_prob
            else:
                return 0,1,1,0,0,0

    def _isoform_differential_usage(self, thisnormdf, thisexp):
        #define major transcript
        #highest mean across time in averaged replicates
    
        #vis
        # for n in range(normdf.shape[1]):
        #     plt.plot(normdf[0,n,:])
        # plt.show()
        
        iso_ratio = np.zeros((thisnormdf.shape[1],thisnormdf.shape[1]))
        iso_diff_value = np.zeros((thisnormdf.shape[1],thisnormdf.shape[1]))
        maj_pval = np.zeros((thisnormdf.shape[1],thisnormdf.shape[1]))
        min_pval = np.zeros((thisnormdf.shape[1],thisnormdf.shape[1]))
        best_switch_point = [[] for _ in range(thisnormdf.shape[1])]
        corr_list = np.zeros((thisnormdf.shape[1],thisnormdf.shape[1]))
        enrichness = np.zeros((thisnormdf.shape[1],thisnormdf.shape[1]))
        iso_prob = np.zeros((thisnormdf.shape[1],thisnormdf.shape[1]))
        all_switch_point = [[] for _ in range(thisnormdf.shape[1])]

        #check for all pairs
        for maj in range(thisnormdf.shape[1]):
            orgarr1 = thisexp[:,maj,:] 
            arr1 = thisnormdf[:,maj,:]
            for x in range(thisnormdf.shape[1]): #each minor transcript
                if x == maj:
                    iso_ratio[maj, x], iso_diff_value[maj, x] = 0, 0
                    best_switch_point[maj].append(0)
                    all_switch_point[maj].append(0)
                    corr_list[maj, x] = 0
                    continue
                arr2 = thisnormdf[:,x,:]
                orgarr2 = thisexp[:,x,:]

                ##finding switch points, correlation
               
                iso_ratio[maj, x], allsp, final_sp, corr_list[maj, x] = self._iso_switch_between_arrays(arr1, arr2, orgarr1, orgarr2)
          
                ##calculate diff. value, p-values
                iso_diff_value[maj, x], maj_pval[maj, x], min_pval[maj, x], bs, enrichness[maj,x], iso_prob[maj, x] = self._diff_before_after_switch(thisnormdf, arr1, arr2, allsp, final_sp)
                
                best_switch_point[maj].append(bs)
                all_switch_point[maj].append(final_sp)


        return iso_ratio, maj_pval, min_pval, iso_diff_value, best_switch_point, corr_list, enrichness, all_switch_point, iso_prob


    def _correlation_coef(self, thisnormdf, maj):
        m_normdf = np.median(thisnormdf, axis=0)
        
        corr_mat = (1-np.corrcoef(m_normdf))/2


        return corr_mat
    

    def _exclusive_domain(self, iso1, iso2):
        if iso1 in self.trans_pfam_dict.keys():
            pfams1 = self.trans_pfam_dict[iso1]
        else: 
            pfams1 = []

        if iso2 in self.trans_pfam_dict.keys():
            pfams2 = self.trans_pfam_dict[iso2]
        else: 
            pfams2 = []
        
        diff = set(pfams1).symmetric_difference(set(pfams2))
        diff = [x for x in list(diff) if not type(x)==float]
        return diff

    def _diff_exons(self, iso1, iso2, final_sp, isoid):
        if iso1 in self.trans_exon_dict.keys():
            exons1 = self.trans_exon_dict[iso1]
        else: 
            exons1 = []

        if iso2 in self.trans_exon_dict.keys():
            exons2 = self.trans_exon_dict[iso2]
        else: 
            exons2 = []

        ##major or minor isoforms
        ##define isoform_from

        iso1_id = np.where(np.isin(self.normdf[isoid]['trans_id'], iso1))[0][0]
        iso2_id = np.where(np.isin(self.normdf[isoid]['trans_id'], iso2))[0][0]
        iso1_values = np.median(self.normdf[isoid]['array'][:,iso1_id, final_sp-1])
        iso2_values = np.median(self.normdf[isoid]['array'][:,iso2_id, final_sp-1])
        
        if iso1_values > iso2_values:
            majoriso, minoriso = iso1, iso2
            majoriso_exons, minoriso_exons = exons1, exons2
        else:
            majoriso, minoriso = iso2, iso1
            majoriso_exons, minoriso_exons = exons2, exons1

        gain_exons = set(majoriso_exons).difference(set(minoriso_exons))
        loss_exons = set(minoriso_exons).difference(set(majoriso_exons))
        nondiff = set(exons1).intersection(set(exons2))
        gain_exons = [x for x in list(gain_exons) if not type(x)==float]
        loss_exons = [x for x in list(loss_exons) if not type(x)==float]
        nondiff = [x for x in list(nondiff) if not type(x)==float]

        return majoriso, minoriso, gain_exons, loss_exons, nondiff
        
                
    def detect_isoform_switch(self, filtering = True, prob_cutoff=0.5, corr_cutoff=0.5, event_im_cutoff = 0.1, min_diff = 0.05, p_val_cutoff = 0.05, n_permutations=100, adjustp='fdr_bh', diff_domain=True):
        """
        Detect isoform switching events along time series.
        
        Parameters
        ----------
        combine : {'median', 'mean'}
            aggregation methods for replicates. Default="median"

        filtering : (boolean, default=True) 
            if True, low expression genes will be filtered out.

        filter_cutoff : default=2
            expression mean cutoff to filter.

        corr_cutoff : default = 0.7
            minimum correlation of isoform pairs to be included in the output.

        p_val_cutoff : default = 0.05   
            significant p-value cutoff to be included in the output.

        min_diff : default = 0.1
            minimum differences of relative abundance to be included in the output.

        event_im_cutoff : default = 0.1
            minimum event importance to be included in the output.

        adjustp : str {'fdr_bh' (default), 'holm_bonf', 'bonf'}
            Method for multiple testing
            bonf: Bonferroni method
            holm_bonf: holm-bonferroni method
            fdr_bh: Benjamin-hochberg false discovery rate

        n_permutations :
            Number of permutations if permutation test is used.

        Return
        ------

        DataFrame 

        """
        #calculate probability and correlation for switching
        #
        #isodf = pd.DataFrame({'symbol':DataSet.symbs, 'isoform_id':DataSet.isoform_id})
        self.min_diff = min_diff
        self.corr_cutoff = corr_cutoff
        self.event_im_cutoff = event_im_cutoff
        self.timeserieslist = self.dataset.timeserieslist
        self.transcript_id = self.dataset.transcript_id
        self.symbs = self.dataset.symbs
        self.gene_id = self.dataset.gene_id
        if filtering==True:
            lowgenes = self._remove_low_expressed()
        else:
            lowgenes=set()
        print(str(len(set(lowgenes))) + " of genes with low expression are filtered out!!.")

        print(f"calculating for {len(self.normdf.keys())} genes...")
              
        iso_pairs_id = defaultdict(list)
        iso_dict_info = defaultdict(tuple)
        for gene, values in self.normdf.items():
            idxs = values['indices']
            thisnormdf = values['normarr']
            thisexp = values['array']

            
            # maj = np.argmax(np.median(np.median(thisexp, axis=0), axis=1))
            # secondmaj = None
            # if thisnormdf.shape[1] > 2:
            #     ##get second maj isoform
            #     secondmaj = np.argsort(np.median(np.median(thisexp, axis=0), axis=1))[-2]

            # self.normdf[gene]['majs'].append(maj)
            # self.normdf[gene]['majs'].append(secondmaj)
            #calculate correlations
            #corrmat= self.correlation_coef(thisnormdf, maj)

            #calculate isoform ratio
            res_maj = self._isoform_differential_usage(thisnormdf, thisexp)

            #return major transcript and transcript that switch
            if self.dataset.reps1 > 1:
                for ids1, iso_ratio0 in enumerate(res_maj[0]):
                    for ids2, iso_ratio in enumerate(iso_ratio0):
                        if iso_ratio >0 :
                            ##
                            majoriso, minoriso, gain_exons, loss_exons, nondiff = self._diff_exons(self.transcript_id[idxs[ids1]], self.transcript_id[idxs[ids2]], res_maj[4][ids1][ids2], str(gene))
                            iso_pairs_id['gene'].append(str(gene))
                            iso_pairs_id['gene_symb'].append(self.symbs[idxs[ids1]])
                            iso_pairs_id['major_transcript'].append(majoriso)
                            iso_pairs_id['minor_transcript'].append(minoriso)
                            iso_pairs_id['switch_prob'].append(res_maj[8][ids1][ids2])
                            iso_pairs_id['diff'].append(res_maj[3][ids1, ids2])
                            iso_pairs_id['maj_pval'].append(res_maj[1][ids1, ids2])
                            iso_pairs_id['min_pval'].append(res_maj[2][ids1, ids2])
                            iso_pairs_id['p_value'].append(self._combine_pvals([res_maj[1][ids1,ids2], res_maj[2][ids1, ids2]]))
                            iso_pairs_id['corr'].append(res_maj[5][ids1, ids2])
                            iso_pairs_id['best_switch_point'].append(res_maj[4][ids1][ids2])
                            iso_pairs_id['event_importance'].append(res_maj[6][ids1][ids2])
                            iso_pairs_id['gain_exons'].append(gain_exons)
                            iso_pairs_id['loss_exons'].append(loss_exons)
                            iso_pairs_id['constant_exons'].append(nondiff)
                            iso_pairs_id['all_switch_points'].append(res_maj[7][ids1][ids2])

                            iso_dict_info[self.transcript_id[idxs[ids1]]] = (gene, ids1)
                            iso_dict_info[self.transcript_id[idxs[ids2]]] = (gene, ids2)

                            if diff_domain == True:
                                iso_pairs_id['exclusive_domains'].append(self._exclusive_domain(self.transcript_id[idxs[ids1]],self.transcript_id[idxs[ids2]]))
                            # iso_dict_info[self.isoform_id[idxs[res_maj[3]]]] = (gene, res_maj[3])
                            # iso_dict_info[self.isoform_id[idxs[ids]]] = (gene, ids)


            else:
                 #p-values is calculated 
                # for ids, iso_ratio in enumerate(res_maj[0]):
                #     if iso_ratio >0 :
                #         iso_pairs_id['gene'].append(gene)
                #         iso_pairs_id['major_transcript'].append(self.transcript_id[idxs[res_maj[3]]])
                #         iso_pairs_id['minor_transcript'].append(self.transcript_id[idxs[ids]])
                #         iso_pairs_id['iso_ratio'].append(iso_ratio)
                #         iso_pairs_id['diff'].append(res_maj[4][ids])
                #         iso_pairs_id['corr'].append(res_maj[6][ids])
                #         iso_pairs_id['final_sp'].append(res_maj[5][ids])
                #         iso_dict_info[self.transcript_id[idxs[res_maj[3]]]] = (gene, res_maj[3])
                #         iso_dict_info[self.transcript_id[idxs[ids]]] = (gene, ids)

                for ids1, iso_ratio0 in enumerate(res_maj[0]):
                    for ids2, iso_ratio in enumerate(iso_ratio0):
                        if iso_ratio >0 :
                            ##
                            majoriso, minoriso, gain_exons, loss_exons, nondiff = self._diff_exons(self.transcript_id[idxs[ids1]], self.transcript_id[idxs[ids2]], res_maj[4][ids1][ids2], str(gene))
                            iso_pairs_id['gene'].append(str(gene))
                            iso_pairs_id['gene_symb'].append(self.symbs[idxs[ids1]])
                            iso_pairs_id['major_transcript'].append(majoriso)
                            iso_pairs_id['minor_transcript'].append(minoriso)
                            iso_pairs_id['switch_prob'].append(res_maj[8][ids1][ids2])
                            iso_pairs_id['diff'].append(res_maj[3][ids1, ids2])
                            iso_pairs_id['corr'].append(res_maj[5][ids1, ids2])
                            iso_pairs_id['best_switch_point'].append(res_maj[4][ids1][ids2])
                            iso_pairs_id['event_importance'].append(res_maj[6][ids1][ids2])
                            iso_pairs_id['gain_exons'].append(gain_exons)
                            iso_pairs_id['loss_exons'].append(loss_exons)
                            iso_pairs_id['constant_exons'].append(nondiff)
                            iso_pairs_id['all_switch_points'].append(res_maj[7][ids1][ids2])

                            iso_dict_info[self.transcript_id[idxs[ids1]]] = (gene, ids1)
                            iso_dict_info[self.transcript_id[idxs[ids2]]] = (gene, ids2)

                            if diff_domain == True:
                                iso_pairs_id['exclusive_domains'].append(self._exclusive_domain(self.transcript_id[idxs[ids1]],self.transcript_id[idxs[ids2]]))

        self.isopairs = iso_pairs_id
        self.iso_dict_info = iso_dict_info
        self.dataset.isoobj = self

        if self.dataset.reps1 == 1:
            cal = basic_pvalue(testobject=self.dataset, object_type="iso_pairs", n_permutations = n_permutations)
            pval = cal.do_calculate() #self iso_function object
            iso_pairs_id['p_value'] = pval

            self.isopairs = iso_pairs_id

        res = pd.DataFrame(iso_pairs_id).sort_values('p_value')
        mct = pvalue_correction(res['p_value'], method=adjustp)
        res['adj_pval'] = mct.corrected_pvals

        res = res[(res['switch_prob']>prob_cutoff) & (res['diff']>min_diff) & (res['adj_pval']<p_val_cutoff) & (res['corr']>corr_cutoff) & (res['event_importance']>event_im_cutoff)]
        res = res.sort_values("switch_prob", ascending=False).drop_duplicates(['major_transcript', 'minor_transcript'])
        print("----Result statistics----")
        print(f"Total genes with IS genes: {res['gene'].unique().shape[0]}")
        print(f"Events found: {res.shape[0]}")
        print(f"Events with affecting domains: {np.sum([1 for x in res[  'exclusive_domains'] if len(x)>0])}")
        print(f"Mean event importance: {res['event_importance'].mean()}")
        print(f"Mean difference before and after switch: {res['diff'].mean()}")
        print("--------------------------")
        print("DONE")
        
        self.is_result = res
        self.dataset.isoobj = self

        return res.reset_index(drop=True)[['gene', 'gene_symb', 'major_transcript', 'minor_transcript', 'switch_prob', 'corr', 'diff', 'event_importance', 'exclusive_domains', 'p_value', 'adj_pval']]


    def _isoform_usage(self, eachgene, norm):
        #eachgene 3D array [replicates, isoforms, timepoints]
        result = np.zeros((eachgene.shape[0], eachgene.shape[1], eachgene.shape[2]-1))
        
        if norm:
            #calculate the size factors 
            sizefactor = np.sum(eachgene, axis=1)
            sizefactor[sizefactor==0] = 1

            #normalize to transcript abundance for each replicates
            normdf = np.array([eachgene[x,:,:] / sizefactor[x] for x in range(eachgene.shape[0])])
        
        else:
            normdf = eachgene
        
        #subtract usage from next time point
        for i in range(eachgene.shape[2]-1):
            result[:,:,i] = np.abs(normdf[:,:,i]-normdf[:,:,i+1])
    
        #sum up and take median for one gene across all replicates
        result = np.median(np.sum(result, axis=1),axis=0)
        
        return result
            

    def total_isoform_usage(self, isd_result=None, norm=True, gene_level=True):
        '''
        only genes with more than 1 isoform are computed.

        Parameters
        ----------
        ids_result 
            the result dataframe of isoform switch detection

        norm : `boolean`, default=True
            if True, it normalizes time series matrix to relative abundance to gene expression.

        gene_level : `boolean`, default=True
            if True, it calculates total isoform usage for each gene, otherwise individual isoform usage for each isoform

        Return
        -------
        New DataSet Object
        
        '''
        isodf = pd.DataFrame({'symbol':self.dataset.gene_id, 'isoform_id':self.dataset.transcript_id})
        groupdict = isodf.groupby(['symbol']).indices
        symbolmap = dict(zip(self.dataset.gene_id, self.dataset.symbs))
        tp = self.dataset.timepts
        
        result_df = pd.DataFrame(columns=['entrez','gene'] + ['tp{}'.format(x) for x in range(self.dataset.timepts-1)])
                
        if isd_result is not None:
            keepgenes = isd_result['gene'].unique().tolist()
        else:
            keepgenes = np.unique(self.dataset.gene_id)  

        if gene_level is True:        
            for gene, iso in groupdict.items():
                if len(iso) > 1 and gene in set(keepgenes):
                # iso_switch = isd_result[isd_result['gene']==gene]
                # all_isos = pd.concat([iso_switch.major_isoform, iso_switch.minor_isoform]).unique().tolist()
                # keepiso = np.where(pd.Series(DataSet.isoform_id).isin(all_isos))[0]

                    eachgene = np.array(self.dataset.timeserieslist[:,iso,:], dtype="double")
                    
                    #calculate isoform usage for each gene
                    res = self._isoform_usage(eachgene, norm)
                
                    result_df.loc[result_df.shape[0]] = [gene, symbolmap[gene]] + res.tolist()

            #make changes in DataSet and return new dataset
            self.dataset.tiu_ts = np.array(result_df.iloc[:,2:])
            self.dataset.tiu_timeserieslist = np.expand_dims(np.array(result_df.iloc[:,2:]), axis=0)
            self.dataset.tiu_gene_id = result_df['entrez']
            self.dataset.tiu_symbs = result_df['gene']
            self.dataset.tiu_timepts = tp-1

        
        elif isd_result is not None and gene_level is False:
            keepgenes = isd_result['gene'].unique().tolist()
            keep_genes_arrays = np.hstack([self.normdf[x]['normarr'] for x in keepgenes])

            result = np.zeros((keep_genes_arrays.shape[0], keep_genes_arrays.shape[1], keep_genes_arrays.shape[2]-1))

            keep_trans_indices = reduce(lambda x,y: x+y, [self.normdf[x]['indices'] for x in keepgenes])
            result_trans = self.dataset.transcript_id[keep_trans_indices]
            result_entrez = self.dataset.gene_id[keep_trans_indices]
            result_sym = self.dataset.symbs[keep_trans_indices]

            for i in range(keep_genes_arrays.shape[2]-1):
                result[:,:,i] = np.abs(keep_genes_arrays[:,:,i]-keep_genes_arrays[:,:,i+1])


            self.dataset.tiu_ts = result
            self.dataset.tiu_timeserieslist = result
            self.dataset.tiu_gene_id = result_entrez
            self.dataset.tiu_symbs = result_sym
            self.dataset.tiu_transcript_id = result_trans
            self.dataset.tiu_timepts = tp-1

        return self.dataset


    def isoform_abundance(self, is_result=None):
        """
        Get isoform relative abundance based on isoform switched genes
        """
        if is_result is None:
            keepgenes = is_result['gene'].unique().tolist()
        else:
            keepgenes = self.dataset.gene_id

        result = np.hstack([self.normdf[x]['normarr'] for x in keepgenes])

        keep_trans_indices = reduce(lambda x,y: x+y, [self.normdf[x]['indices'] for x in keepgenes])
        result_trans = self.dataset.transcript_id[keep_trans_indices]
        result_entrez = self.dataset.gene_id[keep_trans_indices]

        tiuds = self.dataset.__copy__()
        tiuds.ts[0] = result
        tiuds.timeserieslist = result
        tiuds.gene_id = result_entrez
        tiuds.transcript_id = result_trans
        tiuds.timepts = self.dataset.timepts
        tiuds.reps1 = self.dataset.reps1

        return tiuds



import numpy as np
import pandas as pd
from collections import defaultdict

from ._preprocessing._discretization import discretization_with_steps, discretize_timeseries, discretize_replicates

def _map_to_gene_name(keys, keytype, species, transid_type = None):
    keymap = {'entrezgeneid': 'NCBI gene (formerly Entrezgene) ID', 'ensemblgeneid': 'Gene stable ID', 'ensembltransid' : 'Transcript stable ID', 'ensembltransidver': "Transcript stable ID version"}
    ann = pd.read_csv("./spycone_pkg/data/annotation/{}_ann.csv".format(species), index_col=keymap[keytype], dtype=str)
    ann.index = list(map(str, list(map(lambda x: int(x) if not np.isnan(x) else x, ann.index))))
    ann_dict = ann.to_dict()
    
    symbs = []
    for k in keys:
        try:
            symbs.append(ann_dict['Gene name'][str(k)])
        except:
            symbs.append('n')

    return symbs

def _map_trans_to_gene(keys, keytype, species, transid_type = None):
    keymap = {'entrezgeneid': 'NCBI gene (formerly Entrezgene) ID', 'ensemblgeneid': 'Gene stable ID', 'ensembltransid' : 'Transcript stable ID', 'ensembltransidver': "Transcript stable ID version"}
    ann = pd.read_csv("./spycone_pkg/data/annotation/{}_ann.csv".format(species), index_col=keymap[transid_type], dtype=str)
    ann_dict = ann.to_dict()

    a = []
    for k in keys:
        try:
            a.append(str(ann_dict[keymap[keytype]][k]))
        except:
            a.append('n')

    return a

##dataset and settings
class DataSet():
    """ 
    Input dataset.

    Parameters
    ----------

    ts 
        matrix or dataframe of the time series dataset.

    gene_id 
        list of gene id. (id that matched the biological network, default network : entrez ID). If gene_id is not given, entrez gene id will be mapped. 

    transcript_id 
        list of transcript id if the expression matrix is in transcript-level.

    keytype
        type of ID in gene list ('entrezgeneid', 'ensemblgeneid'). If only transcript ID is given, this serves as the keytype that will be mapped to gene_id.
    
    species
        Specifying species ID for annotation (human : 9606, mouse: 10090)

    reps1 
        number of replicates 

    timepts 
        number of time points

    symbs : (optional)
        list of gene symbols or gene names (can be automatically mapped for human and mouse)

    discretization_steps : (optional)
        discretize the expression matrix according to the number of steps given


    Attributes
    -----------
    timeserieslist 
        automatically generated 3D-array for analysis

    Methods
    ----------
    __copy__ 
        copy function for the dataset object

    remove_objects 
        remove the given index of the object

    """
    SPECIES = [9606, 10090]
    KEYTYPE = ['entrezgeneid', 'ensemblgeneid', 'ensembltransid', 'ensembltransidver']
    def __init__(
        self, 
        ts, 
        species,
        keytype,
        reps1,
        timepts,
        gene_id=None,
        transcript_id = None, 
        timeserieslist = None,
        symbs = None,
        discretization_steps = None):
        """
        ts : matrix of the time series data.
        gene_list : list of the gene id
        reps1 : Number of replicates for ts.
        timepts : Number of time points.
        """
        self.ts = [ts]
        self.reps1 = reps1
        self.gene_id = gene_id
        self.species=species
        self.keytype=keytype
        self.timepts = timepts
        self.timeserieslist = []
        self.symbs = symbs
        self.transcript_id = transcript_id
        self.discretization_steps = discretization_steps

        
        #check attributes
        if not hasattr(self.ts, "shape"):
            self.ts[0] = np.array(self.ts[0], dtype="double")
        
        if self.timepts*self.reps1 != self.ts[0].shape[1]:
            raise ValueError("Number of columns is not the same as number of time points.")

        if self.species not in self.SPECIES:
            raise ValueError("Please provide a supported species ID.")

        if self.keytype not in self.KEYTYPE:
            raise ValueError("Please provide a supported key type.")

        ######gene list check
        if self.gene_id is None and self.transcript_id is None:
            raise ValueError("Please provide at least gene ID or transcript ID.")

        TRANS_ID_TYPE = None
        if self.transcript_id is not None:
            if "." in self.transcript_id[0]:
                TRANS_ID_TYPE = "ensembltransidver"
            else:
                TRANS_ID_TYPE = "ensembltransid"

        if self.transcript_id is not None and self.gene_id is None:
            self.gene_id = _map_trans_to_gene(self.transcript_id, self.keytype, self.species, transid_type=TRANS_ID_TYPE)

        if self.symbs is None:
            self.symbs = _map_to_gene_name(self.gene_id, self.keytype, self.species, transid_type=TRANS_ID_TYPE)

        if not isinstance(self.gene_id, list):
            self.gene_id = [str(x) for x in self.gene_id]
            
        if not isinstance(self.symbs, (np.ndarray)):
            self.symbs = np.array(self.symbs)
        # if (self.ts[0].shape[0] != self.gene_list[0] | self.gene_list[0] != self.symbs1[0]).any():
        #     raise ValueError("Input length do not match.") 

        if pd.isna((np.array(self.gene_id))).all():
            raise ValueError("The gene ID list contains NaN values.")

        if pd.isna(self.ts).all():
            raise ValueError("The count matrix contains NaN values.")
        

        tmp = []
        if reps1 > 1:
            tp = self.timepts
            ts = self.ts[0]
            rep = int(self.reps1)

            for r in range(rep):
                start=tp*r
                end = start+tp
                tmp.append(ts[:,start:end])
            
            self.timeserieslist = np.array(tmp, dtype="double")
        else:
            self.timeserieslist = np.expand_dims(ts, axis=0)

        ##discretization
        if self.discretization_steps is not None:
            data_min = np.min(self.timeserieslist)
            data_max = np.max(self.timeserieslist)
            neg = self.discretization_steps[0]
            pos = self.discretization_steps[1]
            discretizeprototype = discretization_with_steps(data_min, data_max, neg, pos)
            

            for i in range(self.timeserieslist.shape[1]):
                tmp_ts = self.timeserieslist[:,i,:]
                discretizepattern = discretize_replicates(tmp_ts, discretizeprototype)
                self.timeserieslist[:,i,:] = tmp_ts 
        
        if self.transcript_id is not None:
            self._get_gene_level()

    @property
    def shape(self):
        return self.timeserieslist.shape

    def __copy__(self):
        return DataSet(ts=self.ts[0],
        gene_id=self.gene_id,
        reps1=self.reps1,
        timepts=self.timepts,
        species=self.species,
        keytype=self.keytype,
        transcript_id=self.transcript_id,
        timeserieslist=self.timeserieslist,
        symbs=self.symbs)

    def _get_gene_level(self):
        genes = defaultdict(list)
        for a in range(len(self.gene_id)):
            genes[self.gene_id[a]].append(a)

        self.genelevel_expression = np.empty(shape=(self.reps1,0,self.timepts))
        self.genelevel_id = []
        self.genelevel_symb = []
        for u,v in genes.items():
            tmp = np.expand_dims(np.sum(self.timeserieslist[:,v,:], axis=1),axis=1)
            self.genelevel_expression = np.concatenate([self.genelevel_expression, tmp], axis=1)
            self.genelevel_id.append(u)
            self.genelevel_symb.append(self.symbs[v[0]])
        
    def _remove_objects(self, objects_to_remove):
        if not isinstance(objects_to_remove, list):
            objects_to_remove = list(objects_to_remove)

        #remove objects from entries
        if self.transcript_id is not None:
            self.transcript_id = np.delete(np.array(self.transcript_id), objects_to_remove)

        self.ts[0] = np.delete(np.array(self.ts[0]), objects_to_remove, axis=0)
        tmp_genelist = np.delete(np.array(self.gene_id), objects_to_remove)
        self.gene_id = tmp_genelist
        tmp_symbs = np.delete(np.array(self.symbs), objects_to_remove)
        self.symbs = tmp_symbs
        tmp_timeseries_id = np.delete(np.arange(np.array(self.timeserieslist).shape[1]), objects_to_remove, axis=0)
        tmp_timeseries = [[self.timeserieslist[x][y] for y in tmp_timeseries_id] for x in range(np.array(self.timeserieslist).shape[0])]
        self.timeserieslist = np.array(tmp_timeseries, dtype="double")

        #recalculate gene level expression
        self._get_gene_level()

    def _add_clustering_object(self, clusterobj):
        self.clusterobj = clusterobj

    def _add_iso_switch_obj(self, isoobj):
        self.isoobj = isoobj

    



import pandas as pd
import numpy as np
import time
import sys


def discretization_with_steps(min, max, neg, pos) -> list:
    '''
    min max of the time series object
    neg = negative steps 
    pos = positive steps
    
    '''
    totalsteps = 1
    if min <0:    
        totalsteps += neg
    if max >0:
        totalsteps += pos
    
    discretizevalues = np.zeros(totalsteps)
    index = 0
    if min <0:
        for i in range(neg):
            discretizevalues[index] = (min + (abs(min)/neg)*i)
            index += 1
    discretizevalues[index] = 0
    index += 1
    if max >0:
        for i in range(pos):
            discretizevalues[index] = (max / pos) * i+1
            index+=1
    
    return np.array(discretizevalues)

def discretize_replicates(tsrep, discretizevalues):
    '''
    Handle the replicates
    '''
    newpattern = np.zeros((tsrep.shape))
    newpattern = [discretize_timeseries(i, discretizevalues) for i in tsrep]

    return newpattern


def discretize_timeseries(tsdata, discretizevalues):
    '''
    Discretize the timeseries data, 
    '''
    pattern = np.array(tsdata, dtype="float")
    newvalue = 0
    for i in range(tsdata.shape[0]):
        mindistance = sys.float_info.max
        for j in range(discretizevalues.shape[0]):
            distance = abs(pattern[i]-discretizevalues[j])
            if mindistance > distance:
                mindistance = distance
                newvalue = discretizevalues[j]
        
        pattern[i] = newvalue
    
    return pattern




import argparse
import os
from .core.domino import main as domino_main
from .core.preprocess_slices import create_slices
from .utils.visualize_modules import visualize_modules
from .constants import *

def main_domino():

    parser = argparse.ArgumentParser(description='args')
    parser.add_argument('-a', '--active_genes_files', dest='active_genes_files', help='/path/to/active_genes_files_1,/path/to/active_genes_files_2', default="examples/tnfa_active_genes_file.txt")
    parser.add_argument('-o', '--output_folder', dest='output_folder', help='/path/to/output', default="examples/output")
    parser.add_argument('-n', '--network_file', dest='network_file', help='/path/to/network file', default="examples/huri.sif")
    parser.add_argument('-s', '--slices_file', dest='slices_file', help='/path/to/slices file', default="examples/huri_slices.txt")
    parser.add_argument('-c', '--use_cache', dest='use_cache', help='true', default="true")
    parser.add_argument('-p', '--parallelization', dest='parallelization', default="1")
    parser.add_argument('-v', '--visualization', dest='visualization', default="true")
    parser.add_argument('-sth', '--slice_threshold', dest='slice_threshold', default="0.3", help='threshold of slices')
    parser.add_argument('-mth', '--module_threshold', dest='module_threshold', default="0.05", help='threshold of putative modules')


    args = parser.parse_args()
    active_genes_files = args.active_genes_files.split(",")
    output_folder = args.output_folder
    network_file = args.network_file
    slices_file = args.slices_file
    slice_threshold = float(args.slice_threshold)
    module_threshold = float(args.module_threshold)
    use_cache = args.use_cache=="true"
    parallelization = int(args.parallelization)
    visualization = args.visualization=="true"

    N_OF_THREADS=parallelization
    USE_CACHE=use_cache

    for cur_ag in active_genes_files:
        G_final_modules=domino_main(active_genes_file=cur_ag, network_file=network_file, slices_file=slices_file, slice_threshold=slice_threshold, module_threshold=module_threshold)
        activity_name=os.path.splitext(os.path.split(cur_ag)[-1])[0]
        report_folder=os.path.join(output_folder,activity_name)
        try:
            os.makedirs(report_folder)
        except:
            pass

        out_file=os.path.join(report_folder, "modules.out")
        open(out_file, 'w+').write("\n".join(['[%s]' % ', '.join(list(m.nodes)) for m in G_final_modules]))
        print(f'{len(G_final_modules)} final modules are reported at {out_file}')
        print(visualization)
        if visualization:
            visualize_modules(os.path.splitext(cur_ag.split('/')[-1])[0], G_final_modules, None, network_file, report_folder)

def main_slicer():

    parser = argparse.ArgumentParser(description='args')
    parser.add_argument('-n', '--network_file', dest='network_file', help='/path/to/network', default="examples/huri.sif")
    parser.add_argument('-o', '--output_file', dest='output_file', default="examples/huri.sif", help='/path/to/output')


    args = parser.parse_args()
    network_file = args.network_file
    output_file = args.output_file
    create_slices(network_file, output_file)




if __name__=="__main__":
    main_slicer()
    main_domino()



import pandas as pd
import numpy as np
import time
import random
import warnings

from ..DataSet import DataSet
from ..BioNetwork import BioNetwork

def tanhnormalizer_scaler(feature, seed=1222):
    scaled_feature = []
    smean = np.mean(feature)
    svar = np.var(feature)
    
    for i in feature:
        scaled_feature.append(0.5 * (np.tanh(0.01 * i-smean)/svar)+1)

    return scaled_feature

